(MODULE FARFALLEN (WITH QOBISCHEME XLIB) (MAIN MAIN))
;;; LaHaShem HaAretz U'Mloah

;;; Farfallen 1.0
;;; Copyright 2012 Purdue University. All rights reserved.

;;; This program is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License
;;; as published by the Free Software Foundation; either version 3
;;; of the License, or (at your option) any later version.

;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.

;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

;;; written by:
;;;    Jeffrey Mark Siskind
;;;    School of Electrical and Computer Engineering
;;;    Purdue University
;;;    465 Northwestern Avenue
;;;    Lafayette IN 47907-2035 USA
;;;    voice: +1 765 496-3197
;;;    FAX:   +1 765 494-6440
;;;    qobi@purdue.edu
;;;    ftp://ftp.ecn.purdue.edu/qobi
;;;    http://engineering.purdue.edu/~qobi
;;;             and
;;;    Alexey Radul
;;;    Hamilton Institute
;;;    National University of Ireland Maynooth
;;;    Co. Kildare
;;;    Ireland
;;;    axofch@gmail.com
;;;             and
;;;    David Rush
;;;    Hamilton Institute
;;;    National University of Ireland Maynooth
;;;    Co. Kildare
;;;    Ireland
;;;    kumoyuki@gmail.com
;;;             and
;;;    Barak A. Pearlmutter
;;;    Hamilton Institute
;;;    National University of Ireland Maynooth
;;;    Co. Kildare
;;;    Ireland
;;;    voice: +353 1 7086100
;;;    FAX:   +353 1 7086269
;;;    barak@cs.nuim.ie
;;;    http://www.bcl.hamilton.ie/~barak/

;;; needs work
;;;  1. statement numbers on block beginnings or ends
;;;  2. special dispensation

(include "QobiScheme.sch")

(set! *program* "farfallen")
(set! *panic?* #t)

;;; Macros

;;; Structures

(define-structure statement number statement)

(define-structure program-statement name)

(define-structure subroutine-statement name variables)

(define-structure function-statement name variables)

(define-structure adf-statement assignments)

(define-structure adr-statement assignments)

(define-structure end-statement kind name assignments)

;;; There may be constraints on the kinds of lvalues allowed. External
;;; statements may not allow subscripts. Dimension statements may require
;;; subscripts. Subscripts may be constrained to be variables.
(define-structure external-statement lvalues)

(define-structure dimension-statement lvalues)

(define-structure integer-statement lvalues)

(define-structure real-statement lvalues)

(define-structure double-precision-statement lvalues)

(define-structure complex-statement lvalues)

(define-structure logical-statement lvalues)

(define-structure continue-statement dummy)

(define-structure assignment-statement lvalue expression)

(define-structure read-statement lvalues)

;;; Can there be expressions instead of lvalues?
(define-structure print-statement lvalues)

(define-structure do-statement
 number assignment-statement expression1 expression2)

(define-structure call-statement name expressions)

(define-structure block
 beginning
 statements
 end
 closure-variables
 declaration-lists
 independent-variables
 dependent-variables)

(define-structure implied-do assignments do-statement)

(define-structure unary operator expression)

(define-structure binary left right)

(define-structure parenthesis expressions)

(define-structure forward name epsilon)

;;; This doesn't need epsilon.
(define-structure tangent name)

;;; This doesn't need epsilon.
(define-structure tangent-bundle name)

(define-structure reverse name epsilon)

;;; This doesn't need epsilon.
(define-structure cotangent name)

;;; This doesn't need epsilon.
(define-structure cotangent-bundle name)

(define-structure global name global)

;;; Variables

(define *try* #f)

(define *line* #f)

;;; Parameters

(define *adifor?* #f)

(define *show-AD-translation?* #f)

(define *show-lifting?* #f)

(define *show-specialization?* #f)

(define *show-tree-shake?* #f)

;;; Procedures

(define (internal-error) (panic "This shouldn't happen"))

(define (syntax-error line)
 (cond
  (*try* (*try* #f))
  (else (display "      ")
	(display *line*)
	(newline)
	(display "      ")
	(display
	 (string-append
	  (make-string (- (string-length *line*) (string-length line)) #\space)
	  "^"))
	(newline)
	(panic "Syntax error"))))

(define (try-internal thunk)
 (let ((try *try*) (succeed? #f))
  (call-with-current-continuation
   (lambda (c)
    (set! *try* c)
    (thunk)
    (set! succeed? #t)))
  (set! *try* try)
  succeed?))

(define-macro try
 (lambda (form expander)
  (expander `(try-internal (lambda () ,(second form))) expander)))

(define (unimplemented) (panic "unimplemented"))

;;; General string processing

(define (commas-without-spaces-between ss)
 (reduce (lambda (s1 s2) (string-append s1 "," s2)) ss ""))

(define (commas-between ss)
 (reduce (lambda (s1 s2) (string-append s1 ", " s2)) ss ""))

(define (spaces-between ss)
 (reduce (lambda (s1 s2) (string-append s1 " " s2)) ss ""))

;;; File parsing

(define (remove-sequence-numbers lines)
 (map (lambda (line)
       (when (> (string-length line) 80)
	(display line)
	(newline)
	(panic "Line longer than 80 columns"))
       (if (> (string-length line) 72) (substring line 0 72) line))
      lines))

(define (remove-blank-lines lines)
 (remove-if (lambda (line)
	     (every (lambda (character) (char=? character #\space))
		    (string->list line)))
	    lines))

(define (remove-comments lines)
 (remove-if (lambda (line)
	     (unless (or (char=? (string-ref line 0) #\space)
			 (char-ci=? (string-ref line 0) #\c))
	      (display line)
	      (newline)
	      (panic "Character other than C or space in column 1"))
	     (char-ci=? (string-ref line 0) #\c))
	    lines))

(define (merge-continuation-lines lines)
 (let loop ((lines lines) (result '()))
  (cond
   ((null? lines) (reverse result))
   ((and (>= (string-length (first lines)) 6)
	 (not (char=? (string-ref (first lines) 5) #\space)))
    (unless (string=? (substring (first lines) 1 4) "    ")
     (display (first lines))
     (newline)
     (panic "Continuation line has nonempty statement number"))
    (when (null? result)
     (display (first lines))
     (newline)
     (panic "First line is a continuation line"))
    (loop
     (rest lines)
     (cons
      (string-append (first result)
		     (substring (first lines) 6 (string-length (first lines))))
      (rest result))))
   (else (loop (rest lines) (cons (first lines) result))))))

(define (statement-numbers lines)
 (map (lambda (line)
       (make-statement
	(if (< (string-length line) 2)
	    ""
	    (substring line 1 (min (string-length line) 5)))
	(statement
	 (if (< (string-length line) 7)
	     ""
	     (substring line 6 (string-length line))))))
      lines))

(define (process-lines lines)
 (statement-numbers
  (merge-continuation-lines
   (remove-comments (remove-blank-lines (remove-sequence-numbers lines))))))

;;; Statement parsing

(define (keyword word)
 ;; Returns a line.
 (lambda (line)
  (cond ((zero? (string-length word)) line)
	((zero? (string-length line)) (syntax-error line))
	;; This assumes that spaces are ignored.
	((char=? (string-ref line 0) #\space)
	 ((keyword word) (substring line 1 (string-length line))))
	((char-ci=? (string-ref word 0) (string-ref line 0))
	 ((keyword (substring word 1 (string-length word)))
	  (substring line 1 (string-length line))))
	(else (syntax-error line)))))

(define (end-of-line line)
 ;; Returns #t
 (cond ((zero? (string-length line)) #t)
       ;; This assumes that spaces are ignored.
       ((char=? (string-ref line 0) #\space)
	(end-of-line (substring line 1 (string-length line))))
       (else (syntax-error line))))

(define (number line)
 ;; Returns a list of a line and a number.
 ;; This doesn't (yet) handle e<n>, e+<n>, e-<n>, d<n>, d+<n>, or d-<n>.
 (let loop ((line line) (number ""))
  (cond ((zero? (string-length line))
	 (if (string=? number "") (syntax-error line) (list line number)))
	;; This assumes that spaces are ignored.
	((char=? (string-ref line 0) #\space)
	 (loop (substring line 1 (string-length line)) number))
	((or (char-numeric? (string-ref line 0))
	     ;; This allows more than one decimal point.
	     ;; It also breaks the separation between numbers and .foo.
	     ;; comparison operators.
	     (char=? (string-ref line 0) #\.))
	 (loop (substring line 1 (string-length line))
	       (string-append number (substring line 0 1))))
	((string=? number "") (syntax-error line))
	(else (list line number)))))

(define (variable line)
 ;; Returns a list of a line and a variable.
 (let loop ((line line) (variable ""))
  (cond ((zero? (string-length line))
	 (cond ((string=? variable "") (syntax-error line))
	       (else (if (char-alphabetic? (string-ref variable 0))
			 (list line variable)
			 (syntax-error line)))))
	;; This assumes that spaces are ignored.
	((char=? (string-ref line 0) #\space)
	 (loop (substring line 1 (string-length line)) variable))
	((char-alphanumeric? (string-ref line 0))
	 (loop (substring line 1 (string-length line))
	       (string-append variable (substring line 0 1))))
	((string=? variable "") (syntax-error line))
	(else (if (char-alphabetic? (string-ref variable 0))
		  (list line variable)
		  (syntax-error line))))))

(define (open-parenthesis line) ((keyword "(") line))

(define (close-parenthesis line) ((keyword ")") line))

(define (comma line) ((keyword ",") line))

(define (comma-separated thing)
 ;; Returns a list of a line and a list of things.
 (lambda (line)
  (if (try (thing line))
      (let* ((line-thing (thing line))
	     (line (first line-thing))
	     (thing1 (second line-thing)))
       (let loop ((line line) (things (list thing1)))
	(if (try (comma line))
	    (let* ((line-thing (thing (comma line)))
		   (line (first line-thing))
		   (thing (second line-thing)))
	     (loop line (append things (list thing))))
	    (list line things))))
      (list line '()))))

(define (expression line)
 ;; Returns a list of a line and an expression.
 ;; This intentionally doesn't maintain precedence.
 ;; It also intentionally overaccepts.
 (cond
  ((try (open-parenthesis line))
   (let* ((line-expressions
	   ((comma-separated expression) (open-parenthesis line)))
	  (line (close-parenthesis (first line-expressions)))
	  (expressions (second line-expressions)))
    (if (try (expression line))
	(let* ((line-expression (expression line))
	       (line (first line-expression))
	       (expression (second line-expression)))
	 (list line (make-binary (make-parenthesis expressions) expression)))
	(list line (make-parenthesis expressions)))))
  ;; ** is handled as **
  ;; Are there any others?
  ((try ((keyword "+") line))
   (let* ((line-expression (expression ((keyword "+") line)))
	  (line (first line-expression))
	  (expression (second line-expression)))
    (list line (make-unary "+" expression))))
  ((try ((keyword "-") line))
   (let* ((line-expression (expression ((keyword "-") line)))
	  (line (first line-expression))
	  (expression (second line-expression)))
    (list line (make-unary "-" expression))))
  ((try ((keyword "*") line))
   (let* ((line-expression (expression ((keyword "*") line)))
	  (line (first line-expression))
	  (expression (second line-expression)))
    (list line (make-unary "*" expression))))
  ((try ((keyword "/") line))
   (let* ((line-expression (expression ((keyword "/") line)))
	  (line (first line-expression))
	  (expression (second line-expression)))
    (list line (make-unary "/" expression))))
  ((try ((keyword ".lt.") line))
   (let* ((line-expression (expression ((keyword ".lt.") line)))
	  (line (first line-expression))
	  (expression (second line-expression)))
    (list line (make-unary ".lt." expression))))
  ((try ((keyword ".le.") line))
   (let* ((line-expression (expression ((keyword ".le.") line)))
	  (line (first line-expression))
	  (expression (second line-expression)))
    (list line (make-unary ".le." expression))))
  ((try ((keyword ".gt.") line))
   (let* ((line-expression (expression ((keyword ".gt.") line)))
	  (line (first line-expression))
	  (expression (second line-expression)))
    (list line (make-unary ".gt." expression))))
  ((try ((keyword ".ge.") line))
   (let* ((line-expression (expression ((keyword ".ge.") line)))
	  (line (first line-expression))
	  (expression (second line-expression)))
    (list line (make-unary ".ge." expression))))
  ((try ((keyword ".eq.") line))
   (let* ((line-expression (expression ((keyword ".eq.") line)))
	  (line (first line-expression))
	  (expression (second line-expression)))
    (list line (make-unary ".eq." expression))))
  ((try ((keyword ".ne.") line))
   (let* ((line-expression (expression ((keyword ".ne.") line)))
	  (line (first line-expression))
	  (expression (second line-expression)))
    (list line (make-unary ".ne." expression))))
  ((try ((keyword ".and.") line))
   (let* ((line-expression (expression ((keyword ".and.") line)))
	  (line (first line-expression))
	  (expression (second line-expression)))
    (list line (make-unary ".and." expression))))
  ((try ((keyword ".or.") line))
   (let* ((line-expression (expression ((keyword ".or.") line)))
	  (line (first line-expression))
	  (expression (second line-expression)))
    (list line (make-unary ".or." expression))))
  ((try ((keyword ".not.") line))
   (let* ((line-expression (expression ((keyword ".not.") line)))
	  (line (first line-expression))
	  (expression (second line-expression)))
    (list line (make-unary ".not." expression))))
  ((try (number line))
   (let* ((line-number (number line))
	  (line (first line-number))
	  (number (second line-number)))
    (if (try (expression line))
	(let* ((line-expression (expression line))
	       (line (first line-expression))
	       (expression (second line-expression)))
	 (list line (make-binary number expression)))
	line-number)))
  ((try (variable line))
   (let* ((line-variable (variable line))
	  (line (first line-variable))
	  (variable (second line-variable)))
    (if (try (expression line))
	(let* ((line-expression (expression line))
	       (line (first line-expression))
	       (expression (second line-expression)))
	 (list line (make-binary variable expression)))
	line-variable)))
  (else (syntax-error line))))

(define (lvalue line)
 ;; Returns a list of a line and an lvalue.
 (let* ((line-expression (expression line))
	(line (first line-expression))
	(expression (second line-expression)))
  (list line expression)))

(define (assignment line)
 ;; Returns a list of a line and an assignment.
 (let* ((line-lvalue (lvalue line))
	(line (first line-lvalue))
	(lvalue (second line-lvalue))
	(line-expression (expression ((keyword "=") line)))
	(line (first line-expression))
	(expression (second line-expression)))
  (list line (make-assignment-statement lvalue expression))))

(define (implied-do line)
 ;; Returns a list of a line and a do statement.
 (let* ((line-assignment (assignment line))
	(line (first line-assignment))
	(assignment (second line-assignment))
	(line-expression (expression (comma line)))
	(line (first line-expression))
	(expression1 (second line-expression)))
  (if (try (comma line))
      (let* ((line-expression (expression (comma line)))
	     (line (first line-expression))
	     (expression2 (second line-expression)))
       (list (close-parenthesis line)
	     (make-do-statement #f assignment expression1 expression2)))
      (list (close-parenthesis line)
	    (make-do-statement #f assignment expression1 #f)))))

(define (optional-implied-do thing)
 (lambda (line)
  (if (try (open-parenthesis line))
      (let loop ((line (open-parenthesis line)) (things '()))
       (if (try (implied-do line))
	   (let* ((line-implied-do (implied-do line))
		  (line (first line-implied-do))
		  (implied-do (second line-implied-do)))
	    (list line (make-implied-do things implied-do)))
	   (let* ((line-thing ((optional-implied-do thing) line))
		  (line (first line-thing))
		  (thing (second line-thing)))
	    (loop (comma line) (append things (list thing))))))
      (thing line))))

(define (statement line)
 ;; Returns a statement.
 (set! *line* line)
 (cond
  ((try ((keyword "program") line))
   (let* ((line-variable (variable ((keyword "program") line)))
	  (line (first line-variable))
	  (variable (second line-variable)))
    (end-of-line line)
    (make-program-statement variable)))
  ((try ((keyword "subroutine") line))
   ;; If there are no parameters can/must the parenthesis be eliminated?
   (let* ((line-variable (variable ((keyword "subroutine") line)))
	  (line (first line-variable))
	  (name (second line-variable))
	  (line (open-parenthesis line))
	  (line-variables ((comma-separated variable) line))
	  (line (first line-variables))
	  (variables (second line-variables)))
    (end-of-line (close-parenthesis line))
    (make-subroutine-statement name variables)))
  ((try ((keyword "function") line))
   ;; If there are no parameters can/must the parenthesis be eliminated?
   (let* ((line-variable (variable ((keyword "function") line)))
	  (line (first line-variable))
	  (name (second line-variable))
	  (line (open-parenthesis line))
	  (line-variables ((comma-separated variable) line))
	  (line (first line-variables))
	  (variables (second line-variables)))
    (end-of-line (close-parenthesis line))
    (make-function-statement name variables)))
  ((try ((keyword "adf") line))
   ;; We don't (yet) handle the special dispensation.
   (let* ((line-assignments
	   ((comma-separated (optional-implied-do assignment))
	    (open-parenthesis ((keyword "adf") line))))
	  (line (first line-assignments))
	  (assignments (second line-assignments)))
    (end-of-line (close-parenthesis line))
    (make-adf-statement assignments)))
  ((try ((keyword "adr") line))
   ;; We don't (yet) handle the special dispensation.
   (let* ((line-assignments ((comma-separated (optional-implied-do assignment))
			     (open-parenthesis ((keyword "adr") line))))
	  (line (first line-assignments))
	  (assignments (second line-assignments)))
    (end-of-line (close-parenthesis line))
    (make-adr-statement assignments)))
  ((try ((keyword "end") line))
   (let ((line ((keyword "end") line)))
    (cond
     ((try ((keyword "program") line))
      (let ((line ((keyword "program") line)))
       (if (try (end-of-line line))
	   (make-end-statement 'program #f #f)
	   (let* ((line-variable (variable line))
		  (line (first line-variable))
		  (variable (second line-variable)))
	    (end-of-line line)
	    (make-end-statement 'program variable #f)))))
     ((try ((keyword "subroutine") line))
      (let ((line ((keyword "subroutine") line)))
       (if (try (end-of-line line))
	   (make-end-statement 'subroutine #f #f)
	   (let* ((line-variable (variable line))
		  (line (first line-variable))
		  (variable (second line-variable)))
	    (end-of-line line)
	    (make-end-statement 'subroutine variable #f)))))
     ((try ((keyword "function") line))
      (let ((line ((keyword "function") line)))
       (if (try (end-of-line line))
	   (make-end-statement 'function #f #f)
	   (let* ((line-variable (variable line))
		  (line (first line-variable))
		  (variable (second line-variable)))
	    (end-of-line line)
	    (make-end-statement 'function variable #f)))))
     ((try ((keyword "adf") line))
      (let* ((line-assignments
	      ((comma-separated (optional-implied-do assignment))
	       (open-parenthesis ((keyword "adf") line))))
	     (line (first line-assignments))
	     (assignments (second line-assignments)))
       (end-of-line (close-parenthesis line))
       (make-end-statement 'adf #f assignments)))
     ((try ((keyword "adr") line))
      (let* ((line-assignments
	      ((comma-separated (optional-implied-do assignment))
	       (open-parenthesis ((keyword "adr") line))))
	     (line (first line-assignments))
	     (assignments (second line-assignments)))
       (end-of-line (close-parenthesis line))
       (make-end-statement 'adr #f assignments)))
     ;; We don't handle the case of an unnamed end statement with closing
     ;; assignments anymore because we can't determine whether to look for
     ;; tangent or cotangent closing assignments.
     ((try (end-of-line line)) (make-end-statement #f #f #f))
     (else (syntax-error line)))))
  ((try ((keyword "external") line))
   (let* ((line-lvalues ((comma-separated lvalue) ((keyword "external") line)))
	  (line (first line-lvalues))
	  (lvalues (second line-lvalues)))
    (end-of-line line)
    (make-external-statement lvalues)))
  ((try ((keyword "dimension") line))
   (let* ((line-lvalues ((comma-separated lvalue) ((keyword "dimension") line)))
	  (line (first line-lvalues))
	  (lvalues (second line-lvalues)))
    (end-of-line line)
    (make-dimension-statement lvalues)))
  ((try ((keyword "integer") line))
   (let* ((line-lvalues ((comma-separated lvalue) ((keyword "integer") line)))
	  (line (first line-lvalues))
	  (lvalues (second line-lvalues)))
    (end-of-line line)
    (make-integer-statement lvalues)))
  ((try ((keyword "real") line))
   (let* ((line-lvalues ((comma-separated lvalue) ((keyword "real") line)))
	  (line (first line-lvalues))
	  (lvalues (second line-lvalues)))
    (end-of-line line)
    (make-real-statement lvalues)))
  ((try ((keyword "doubleprecision") line))
   (let* ((line-lvalues
	   ((comma-separated lvalue) ((keyword "doubleprecision") line)))
	  (line (first line-lvalues))
	  (lvalues (second line-lvalues)))
    (end-of-line line)
    (make-double-precision-statement lvalues)))
  ((try ((keyword "complex") line))
   (let* ((line-lvalues ((comma-separated lvalue) ((keyword "complex") line)))
	  (line (first line-lvalues))
	  (lvalues (second line-lvalues)))
    (end-of-line line)
    (make-complex-statement lvalues)))
  ((try ((keyword "logical") line))
   (let* ((line-lvalues ((comma-separated lvalue) ((keyword "logical") line)))
	  (line (first line-lvalues))
	  (lvalues (second line-lvalues)))
    (end-of-line line)
    (make-logical-statement lvalues)))
  ((try ((keyword "continue") line))
   (end-of-line ((keyword "continue") line))
   (make-continue-statement #f))
  ;; This is a hack to parse the *,
  ((try ((keyword "read*,") line))
   (let* ((line-lvalues ((comma-separated lvalue) ((keyword "read*,") line)))
	  (line (first line-lvalues))
	  (lvalues (second line-lvalues)))
    (end-of-line line)
    (make-read-statement lvalues)))
  ;; This is a hack to parse the *,
  ((try ((keyword "print*,") line))
   (let* ((line-lvalues ((comma-separated lvalue) ((keyword "print*,") line)))
	  (line (first line-lvalues))
	  (lvalues (second line-lvalues)))
    (end-of-line line)
    (make-print-statement lvalues)))
  ((try ((keyword "do") line))
   (let* ((line-number (number ((keyword "do") line)))
	  (line (first line-number))
	  (number (second line-number))
	  (line-assignment (assignment line))
	  (line (first line-assignment))
	  (assignment (second line-assignment))
	  (line-expression (expression (comma line)))
	  (line (first line-expression))
	  (expression1 (second line-expression)))
    (cond ((try (comma line))
	   (let* ((line-expression (expression (comma line)))
		  (line (first line-expression))
		  (expression2 (second line-expression)))
	    (end-of-line line)
	    (make-do-statement number assignment expression1 expression2)))
	  (else (end-of-line line)
		(make-do-statement number assignment expression1 #f)))))
  ((try ((keyword "call") line))
   ;; If there are no arguments can/must the parenthesis be eliminated?
   ;; Ditto for a function call.
   (let* ((line-variable (variable ((keyword "call") line)))
	  (line (first line-variable))
	  (variable (second line-variable))
	  (line (open-parenthesis line))
	  (line-expressions ((comma-separated expression) line))
	  (line (first line-expressions))
	  (expressions (second line-expressions)))
    (end-of-line (close-parenthesis line))
    (make-call-statement variable expressions)))
  (else (let* ((line-assignment (assignment line))
	       (line (first line-assignment))
	       (assignment (second line-assignment)))
	 (end-of-line line)
	 assignment))))

;;; Parser

(define (get-block statements beginning)
 (let loop ((statements statements) (result '()))
  (when (null? statements) (panic "Missing END"))
  (let ((statement (first statements)) (statements (rest statements)))
   (cond
    ((or (subroutine-statement? (statement-statement statement))
	 (function-statement? (statement-statement statement))
	 (adf-statement? (statement-statement statement))
	 (adr-statement? (statement-statement statement)))
     (let ((block-statements (get-block statements statement)))
      (loop (second block-statements) (cons (first block-statements) result))))
    ((end-statement? (statement-statement statement))
     (list (make-block beginning (reverse result) statement #f #f #f #f)
	   statements))
    (else (loop statements (cons statement result)))))))

(define (get-blocks statements)
 (let loop ((statements statements) (result '()))
  (if (null? statements)
      (reverse result)
      (let ((statement (first statements)) (statements (rest statements)))
       (cond ((or (program-statement? (statement-statement statement))
		  (subroutine-statement? (statement-statement statement))
		  (function-statement? (statement-statement statement)))
	      (let ((block-statements (get-block statements statement)))
	       (loop (second block-statements)
		     (cons (first block-statements) result))))
	     (else (panic "Not a subprogram")))))))

;;; These can be made more robust.

(define (suffix-name name suffix)
 ;; needs work: to only use during unparse
 (cond
  ((forward? name) (make-forward (suffix-name (forward-name name) suffix) #f))
  ((reverse? name) (make-reverse (suffix-name (reverse-name name) suffix) #f))
  ;; We changed the order to keep the first character the same for implicit
  ;; type declarations. (Note that Tapenade's use of suffix is better than
  ;; Adifor's use of prefix.)
  (suffix (string-append name "_" suffix))
  (else name)))

(define (adf-name index)
 ;; needs work: to only use during unparse
 (string-append "adf_"
		(reduce (lambda (s1 s2) (string-append s1 "_" s2))
			(map number->string (reverse index))
			"")))

(define (adr-name index)
 ;; needs work: to only use during unparse
 (string-append "adr_"
		(reduce (lambda (s1 s2) (string-append s1 "_" s2))
			(map number->string (reverse index))
			"")))

(define (tangent-variable-name name)
 ;; This doesn't need epsilon.
 (string-append name "_tangent"))

(define (cotangent-variable-name name)
 ;; This doesn't need epsilon.
 (string-append name "_cotangent"))

(define (specialize-name name n)
 ;; needs work: to only use during unparse
 (string-append name "_" (number->string n)))

(define (adifor-prefix i) (string (integer->char (+ (char->integer #\a) i))))

(define (tapenade-forward-suffix i)
 (string-append "forward" (number->string i)))

(define (tapenade-reverse-suffix i)
 (string-append "reverse" (number->string i)))

(define (forwardize-name forward)
 (if *adifor?*
     (if (forward-epsilon forward)
	 (string-append
	  (adifor-prefix (forward-epsilon forward)) "_" (forward-name forward))
	 (string-append (forward-name forward) "_forward"))
     (string-append (forward-name forward)
		    "_"
		    (if (forward-epsilon forward)
			(tapenade-forward-suffix (forward-epsilon forward))
			"forward"))))

(define (reverseize-name reverse)
 (when *adifor?* (internal-error))
 (string-append (reverse-name reverse)
		"_"
		(if (reverse-epsilon reverse)
		    (tapenade-reverse-suffix (reverse-epsilon reverse))
		    "reverse")))

(define (closure-variable-name name suffix)
 ;; We changed the order to keep the first character the same for implicit
 ;; type declarations.
 ;; needs work: to only use during unparse
 (string-append name "_" suffix))

;;; Utilities

(define (program-block? statement)
 (and (block? statement)
      (program-statement? (statement-statement (block-beginning statement)))))

(define (subroutine-block? statement)
 (and (block? statement)
      (subroutine-statement?
       (statement-statement (block-beginning statement)))))

(define (function-block? statement)
 (and (block? statement)
      (function-statement? (statement-statement (block-beginning statement)))))

(define (adf-block? statement)
 (and (block? statement)
      (adf-statement? (statement-statement (block-beginning statement)))))

(define (adr-block? statement)
 (and (block? statement)
      (adr-statement? (statement-statement (block-beginning statement)))))

(define (subprogram-name block)
 (let ((statement (statement-statement (block-beginning block))))
  (cond
   ((program-statement? statement) (program-statement-name statement))
   ((subroutine-statement? statement) (subroutine-statement-name statement))
   ((function-statement? statement) (function-statement-name statement))
   (else (internal-error)))))

(define (subprogram-parameters block)
 (let ((statement (statement-statement (block-beginning block))))
  (cond ((program-statement? statement) '())
	((subroutine-statement? statement)
	 (subroutine-statement-variables statement))
	((function-statement? statement)
	 (function-statement-variables statement))
	(else (internal-error)))))

(define (tangents expression)
 (cond
  ((unary? expression) (tangents (unary-expression expression)))
  ((binary? expression)
   (union
    (if (and (string? (binary-left expression))
	     (string-ci=? (binary-left expression) "tangent")
	     (parenthesis? (binary-right expression))
	     (= (length (parenthesis-expressions (binary-right expression)))
		1))
	(let ((expression
	       (first
		(parenthesis-expressions (binary-right expression)))))
	 (cond ((string? expression) (list expression))
	       ((and (binary? expression)
		     (string? (binary-left expression))
		     (parenthesis? (binary-right expression))
		     (>= (length
			  (parenthesis-expressions (binary-right expression)))
			 1))
		(list (binary-left expression)))
	       (else '())))
	'())
    (union (tangents (binary-left expression))
	   (tangents (binary-right expression)))))
  ((parenthesis? expression)
   (map-reduce union '() tangents (parenthesis-expressions expression)))
  ((forward? expression) (internal-error))
  ((tangent? expression) (internal-error))
  ((tangent-bundle? expression) (internal-error))
  ((reverse? expression) (internal-error))
  ((cotangent? expression) (internal-error))
  ((cotangent-bundle? expression) (internal-error))
  ((global? expression) (internal-error))
  ((string? expression) '())
  (else (internal-error))))

(define (cotangents expression)
 (cond
  ((unary? expression) (cotangents (unary-expression expression)))
  ((binary? expression)
   (union
    (if (and (string? (binary-left expression))
	     (string-ci=? (binary-left expression) "cotangent")
	     (parenthesis? (binary-right expression))
	     (= (length (parenthesis-expressions (binary-right expression)))
		1))
	(let ((expression
	       (first
		(parenthesis-expressions (binary-right expression)))))
	 (cond ((string? expression) (list expression))
	       ((and (binary? expression)
		     (string? (binary-left expression))
		     (parenthesis? (binary-right expression))
		     (>= (length
			  (parenthesis-expressions (binary-right expression)))
			 1))
		(list (binary-left expression)))
	       (else '())))
	'())
    (union (cotangents (binary-left expression))
	   (cotangents (binary-right expression)))))
  ((parenthesis? expression)
   (map-reduce union '() cotangents (parenthesis-expressions expression)))
  ((forward? expression) (internal-error))
  ((tangent? expression) (internal-error))
  ((tangent-bundle? expression) (internal-error))
  ((reverse? expression) (internal-error))
  ((cotangent? expression) (internal-error))
  ((cotangent-bundle? expression) (internal-error))
  ((global? expression) (internal-error))
  ((string? expression) '())
  (else (internal-error))))

(define (adf-independent-variables block)
 (map-reduce
  union
  '()
  (lambda (assignment)
   (let loop ((assignment assignment))
    (if (implied-do? assignment)
	(map-reduce union '() loop (implied-do-assignments assignment))
	(tangents (assignment-statement-lvalue assignment)))))
  (adf-statement-assignments (statement-statement (block-beginning block)))))

(define (adf-dependent-variables block)
 (map-reduce
  union
  '()
  (lambda (assignment)
   (let loop ((assignment assignment))
    (if (implied-do? assignment)
	(map-reduce union '() loop (implied-do-assignments assignment))
	(tangents (assignment-statement-expression assignment)))))
  (end-statement-assignments (statement-statement (block-end block)))))

(define (adr-independent-variables block)
 (map-reduce
  union
  '()
  (lambda (assignment)
   (let loop ((assignment assignment))
    (if (implied-do? assignment)
	(map-reduce union '() loop (implied-do-assignments assignment))
	(cotangents (assignment-statement-expression assignment)))))
  (end-statement-assignments (statement-statement (block-end block)))))

(define (adr-dependent-variables block)
 (map-reduce
  union
  '()
  (lambda (assignment)
   (let loop ((assignment assignment))
    (if (implied-do? assignment)
	(map-reduce union '() loop (implied-do-assignments assignment))
	(cotangents (assignment-statement-lvalue assignment)))))
  (adr-statement-assignments (statement-statement (block-beginning block)))))

(define (bound-variables block)
 (cond
  ((program-block? block) '())
  ((subroutine-block? block)
   (subroutine-statement-variables
    (statement-statement (block-beginning block))))
  ((function-block? block)
   (function-statement-variables (statement-statement (block-beginning block))))
  ((adf-block? block)
   (union (adf-independent-variables block)
	  (adf-dependent-variables block)))
  ((adr-block? block)
   (union (adr-independent-variables block)
	  (adr-dependent-variables block)))
  (else (internal-error))))

(define (no-statement-number statement) (make-statement "    " statement))

;;; Translate AD

(define (lvalue-variable lvalue)
 (cond
  ((unary? lvalue) (internal-error))
  ((binary? lvalue)
   (if (and (or (string? (binary-left lvalue))
		(tangent? (binary-left lvalue))
		(cotangent? (binary-left lvalue)))
	    (parenthesis? (binary-right lvalue))
	    (>= (length (parenthesis-expressions (binary-right lvalue))) 1))
       (binary-left lvalue)
       (internal-error)))
  ((parenthesis? lvalue) (internal-error))
  ((forward? lvalue) (internal-error))
  ((tangent? lvalue) lvalue)
  ((tangent-bundle? lvalue) (internal-error))
  ((reverse? lvalue) (internal-error))
  ((cotangent? lvalue) lvalue)
  ((cotangent-bundle? lvalue) (internal-error))
  ((global? lvalue) (internal-error))
  ((string? lvalue) lvalue)
  (else (internal-error))))

(define (tangentify-lvalue lvalue)
 (cond
  ((unary? lvalue) (internal-error))
  ((binary? lvalue)
   (if (and (or (string? (binary-left lvalue))
		(tangent? (binary-left lvalue))
		(cotangent? (binary-left lvalue)))
	    (parenthesis? (binary-right lvalue))
	    (>= (length (parenthesis-expressions (binary-right lvalue))) 1))
       (make-binary (make-tangent (binary-left lvalue)) (binary-right lvalue))
       (internal-error)))
  ((parenthesis? lvalue) (internal-error))
  ((forward? lvalue) (internal-error))
  ((tangent? lvalue) (internal-error))
  ((tangent-bundle? lvalue) (internal-error))
  ((reverse? lvalue) (internal-error))
  ((cotangent? lvalue) (internal-error))
  ((cotangent-bundle? lvalue) (internal-error))
  ((global? lvalue) (internal-error))
  ((string? lvalue) (make-tangent lvalue))
  (else (internal-error))))

(define (cotangentify-lvalue lvalue)
 (cond
  ((unary? lvalue) (internal-error))
  ((binary? lvalue)
   (if (and (string? (binary-left lvalue))
	    (parenthesis? (binary-right lvalue))
	    (>= (length (parenthesis-expressions (binary-right lvalue))) 1))
       (make-binary (make-cotangent (binary-left lvalue)) (binary-right lvalue))
       (internal-error)))
  ((parenthesis? lvalue) (internal-error))
  ((forward? lvalue) (internal-error))
  ((tangent? lvalue) (internal-error))
  ((tangent-bundle? lvalue) (internal-error))
  ((reverse? lvalue) (internal-error))
  ((cotangent? lvalue) (internal-error))
  ((cotangent-bundle? lvalue) (internal-error))
  ((global? lvalue) (internal-error))
  ((string? lvalue) (make-cotangent lvalue))
  (else (internal-error))))

(define (tangentify-expression expression)
 (cond
  ((unary? expression)
   (make-unary (unary-operator expression)
	       (tangentify-expression (unary-expression expression))))
  ((binary? expression)
   (let ((left (tangentify-expression (binary-left expression)))
	 (right (tangentify-expression (binary-right expression))))
    (if (and (string? left)
	     (string-ci=? left "tangent")
	     (parenthesis? right)
	     (= (length (parenthesis-expressions right)) 1))
	(let ((expression (first (parenthesis-expressions right))))
	 (cond ((string? expression) (make-tangent expression))
	       ((and (binary? expression)
		     (string? (binary-left expression))
		     (parenthesis? (binary-right expression))
		     (>= (length
			  (parenthesis-expressions (binary-right expression)))
			 1))
		(make-binary (make-tangent (binary-left expression))
			     (binary-right expression)))
	       (else (make-binary left right))))
	(make-binary left right))))
  ((parenthesis? expression)
   (make-parenthesis
    (map tangentify-expression (parenthesis-expressions expression))))
  ((forward? expression) (internal-error))
  ((tangent? expression) (internal-error))
  ((tangent-bundle? expression) (internal-error))
  ((reverse? expression) (internal-error))
  ((cotangent? expression) (internal-error))
  ((cotangent-bundle? expression) (internal-error))
  ((global? expression) (internal-error))
  ((string? expression) expression)
  (else (internal-error))))

(define (cotangentify-expression expression)
 (cond
  ((unary? expression)
   (make-unary (unary-operator expression)
	       (cotangentify-expression (unary-expression expression))))
  ((binary? expression)
   (let ((left (cotangentify-expression (binary-left expression)))
	 (right (cotangentify-expression (binary-right expression))))
    (if (and (string? left)
	     (string-ci=? left "cotangent")
	     (parenthesis? right)
	     (= (length (parenthesis-expressions right)) 1))
	(let ((expression (first (parenthesis-expressions right))))
	 (cond ((string? expression) (make-cotangent expression))
	       ((and (binary? expression)
		     (string? (binary-left expression))
		     (parenthesis? (binary-right expression))
		     (>= (length
			  (parenthesis-expressions (binary-right expression)))
			 1))
		(make-binary (make-cotangent (binary-left expression))
			     (binary-right expression)))
	       (else (make-binary left right))))
	(make-binary left right))))
  ((parenthesis? expression)
   (make-parenthesis
    (map cotangentify-expression (parenthesis-expressions expression))))
  ((forward? expression) (internal-error))
  ((tangent? expression) (internal-error))
  ((tangent-bundle? expression) (internal-error))
  ((reverse? expression) (internal-error))
  ((cotangent? expression) (internal-error))
  ((cotangent-bundle? expression) (internal-error))
  ((global? expression) (internal-error))
  ((string? expression) expression)
  (else (internal-error))))

(define (translate-adf-opening-assignments block)
 (map-reduce
  append
  '()
  (lambda (assignment)
   (let loop ((assignment assignment))
    (if (implied-do? assignment)
	(append
	 (list (no-statement-number (implied-do-do-statement assignment)))
	 (map-reduce append '() loop (implied-do-assignments assignment))
	 (list (no-statement-number (make-end-statement 'do #f #f))))
	(list (no-statement-number
	       (make-assignment-statement
		(tangentify-expression (assignment-statement-lvalue assignment))
		(tangentify-expression
		 (assignment-statement-expression assignment))))))))
  (adf-statement-assignments (statement-statement (block-beginning block)))))

(define (translate-adf-closing-assignments block)
 (map-reduce
  append
  '()
  (lambda (assignment)
   (let loop ((assignment assignment))
    (if (implied-do? assignment)
	(append
	 (list (no-statement-number (implied-do-do-statement assignment)))
	 (map-reduce append '() loop (implied-do-assignments assignment))
	 (list (no-statement-number (make-end-statement 'do #f #f))))
	(list (no-statement-number
	       (make-assignment-statement
		(tangentify-expression (assignment-statement-lvalue assignment))
		(tangentify-expression
		 (assignment-statement-expression assignment))))))))
  (end-statement-assignments (statement-statement (block-end block)))))

(define (translate-adr-opening-assignments block)
 (map-reduce
  append
  '()
  (lambda (assignment)
   (let loop ((assignment assignment))
    (if (implied-do? assignment)
	(append
	 (list (no-statement-number (implied-do-do-statement assignment)))
	 (map-reduce append '() loop (implied-do-assignments assignment))
	 (list (no-statement-number (make-end-statement 'do #f #f))))
	(list
	 (no-statement-number
	  (make-assignment-statement
	   (cotangentify-expression (assignment-statement-lvalue assignment))
	   (cotangentify-expression
	    (assignment-statement-expression assignment))))))))
  (adr-statement-assignments (statement-statement (block-beginning block)))))

(define (translate-adr-closing-assignments block)
 (map-reduce
  append
  '()
  (lambda (assignment)
   (let loop ((assignment assignment))
    (if (implied-do? assignment)
	(append
	 (list (no-statement-number (implied-do-do-statement assignment)))
	 (map-reduce append '() loop (implied-do-assignments assignment))
	 (list (no-statement-number (make-end-statement 'do #f #f))))
	(list
	 (no-statement-number
	  (make-assignment-statement
	   (cotangentify-expression (assignment-statement-lvalue assignment))
	   (cotangentify-expression
	    (assignment-statement-expression assignment))))))))
  (end-statement-assignments (statement-statement (block-end block)))))

(define (old-declarations-for variable block)
 (removeq
  #f
  (map
   (lambda (statement)
    (if (block? statement)
	#f
	(let ((statement (statement-statement statement)))
	 (cond
	  ((external-statement? statement)
	   (let ((lvalues
		  (remove-if-not
		   (lambda (lvalue) (equal? variable (lvalue-variable lvalue)))
		   (external-statement-lvalues statement))))
	    (if (null? lvalues) #f (make-external-statement lvalues))))
	  ((dimension-statement? statement)
	   (let ((lvalues
		  (remove-if-not
		   (lambda (lvalue) (equal? variable (lvalue-variable lvalue)))
		   (dimension-statement-lvalues statement))))
	    (if (null? lvalues) #f (make-dimension-statement lvalues))))
	  ((integer-statement? statement)
	   (let ((lvalues
		  (remove-if-not
		   (lambda (lvalue) (equal? variable (lvalue-variable lvalue)))
		   (integer-statement-lvalues statement))))
	    (if (null? lvalues) #f (make-integer-statement lvalues))))
	  ((real-statement? statement)
	   (let ((lvalues
		  (remove-if-not
		   (lambda (lvalue) (equal? variable (lvalue-variable lvalue)))
		   (real-statement-lvalues statement))))
	    (if (null? lvalues) #f (make-real-statement lvalues))))
	  ((double-precision-statement? statement)
	   (let ((lvalues
		  (remove-if-not
		   (lambda (lvalue) (equal? variable (lvalue-variable lvalue)))
		   (double-precision-statement-lvalues statement))))
	    (if (null? lvalues) #f (make-double-precision-statement lvalues))))
	  ((complex-statement? statement)
	   (let ((lvalues
		  (remove-if-not
		   (lambda (lvalue) (equal? variable (lvalue-variable lvalue)))
		   (complex-statement-lvalues statement))))
	    (if (null? lvalues) #f (make-complex-statement lvalues))))
	  ((logical-statement? statement)
	   (let ((lvalues
		  (remove-if-not
		   (lambda (lvalue) (equal? variable (lvalue-variable lvalue)))
		   (logical-statement-lvalues statement))))
	    (if (null? lvalues) #f (make-logical-statement lvalues))))
	  (else #f)))))
   (block-statements block))))

(define (declarations-for variable context)
 (let loop ((context context))
  (if (null? context)
      '()
      (let* ((block (first context))
	     (declarations
	      (removeq
	       #f
	       (map
		(lambda (statement)
		 (if (block? statement)
		     #f
		     (let ((statement (statement-statement statement)))
		      (cond
		       ((external-statement? statement)
			(let ((lvalues
			       (remove-if-not
				(lambda (lvalue)
				 (equal? variable (lvalue-variable lvalue)))
				(external-statement-lvalues statement))))
			 (if (null? lvalues)
			     #f
			     (make-external-statement lvalues))))
		       ((dimension-statement? statement)
			(let ((lvalues
			       (remove-if-not
				(lambda (lvalue)
				 (equal? variable (lvalue-variable lvalue)))
				(dimension-statement-lvalues statement))))
			 (if (null? lvalues)
			     #f
			     (make-dimension-statement lvalues))))
		       ((integer-statement? statement)
			(let ((lvalues
			       (remove-if-not
				(lambda (lvalue)
				 (equal? variable (lvalue-variable lvalue)))
				(integer-statement-lvalues statement))))
			 (if (null? lvalues)
			     #f
			     (make-integer-statement lvalues))))
		       ((real-statement? statement)
			(let ((lvalues
			       (remove-if-not
				(lambda (lvalue)
				 (equal? variable (lvalue-variable lvalue)))
				(real-statement-lvalues statement))))
			 (if (null? lvalues) #f (make-real-statement lvalues))))
		       ((double-precision-statement? statement)
			(let ((lvalues
			       (remove-if-not
				(lambda (lvalue)
				 (equal? variable (lvalue-variable lvalue)))
				(double-precision-statement-lvalues
				 statement))))
			 (if (null? lvalues)
			     #f
			     (make-double-precision-statement lvalues))))
		       ((complex-statement? statement)
			(let ((lvalues
			       (remove-if-not
				(lambda (lvalue)
				 (equal? variable (lvalue-variable lvalue)))
				(complex-statement-lvalues statement))))
			 (if (null? lvalues)
			     #f
			     (make-complex-statement lvalues))))
		       ((logical-statement? statement)
			(let ((lvalues
			       (remove-if-not
				(lambda (lvalue)
				 (equal? variable (lvalue-variable lvalue)))
				(logical-statement-lvalues statement))))
			 (if (null? lvalues)
			     #f
			     (make-logical-statement lvalues))))
		       (else #f)))))
		(block-statements block)))))
       (if (null? declarations)
	   (if (and (or (subroutine-block? block)
			(function-block? block))
		    (member variable (bound-variables block)))
	       '()
	       (loop (rest context)))
	   declarations)))))

(define (tangent-declarations-for variable context)
 (let loop ((context context))
  (if (null? context)
      '()
      (let* ((block (first context))
	     (declarations
	      (removeq
	       #f
	       (map
		(lambda (statement)
		 (if (block? statement)
		     #f
		     (let ((statement (statement-statement statement)))
		      (cond
		       ((external-statement? statement)
			(let ((lvalues
			       (map tangentify-lvalue
				    (remove-if-not
				     (lambda (lvalue)
				      (equal? variable
					      (lvalue-variable lvalue)))
				     (external-statement-lvalues statement)))))
			 (if (null? lvalues)
			     #f
			     (make-external-statement lvalues))))
		       ((dimension-statement? statement)
			(let ((lvalues
			       (map tangentify-lvalue
				    (remove-if-not
				     (lambda (lvalue)
				      (equal? variable
					      (lvalue-variable lvalue)))
				     (dimension-statement-lvalues statement)))))
			 (if (null? lvalues)
			     #f
			     (make-dimension-statement lvalues))))
		       ((integer-statement? statement)
			(let ((lvalues
			       (map tangentify-lvalue
				    (remove-if-not
				     (lambda (lvalue)
				      (equal? variable
					      (lvalue-variable lvalue)))
				     (integer-statement-lvalues statement)))))
			 (if (null? lvalues)
			     #f
			     (make-integer-statement lvalues))))
		       ((real-statement? statement)
			(let ((lvalues
			       (map tangentify-lvalue
				    (remove-if-not
				     (lambda (lvalue)
				      (equal? variable
					      (lvalue-variable lvalue)))
				     (real-statement-lvalues statement)))))
			 (if (null? lvalues) #f (make-real-statement lvalues))))
		       ((double-precision-statement? statement)
			(let ((lvalues
			       (map tangentify-lvalue
				    (remove-if-not
				     (lambda (lvalue)
				      (equal? variable
					      (lvalue-variable lvalue)))
				     (double-precision-statement-lvalues
				      statement)))))
			 (if (null? lvalues)
			     #f
			     (make-double-precision-statement lvalues))))
		       ((complex-statement? statement)
			(let ((lvalues
			       (map tangentify-lvalue
				    (remove-if-not
				     (lambda (lvalue)
				      (equal? variable
					      (lvalue-variable lvalue)))
				     (complex-statement-lvalues statement)))))
			 (if (null? lvalues)
			     #f
			     (make-complex-statement lvalues))))
		       ((logical-statement? statement)
			(let ((lvalues
			       (map tangentify-lvalue
				    (remove-if-not
				     (lambda (lvalue)
				      (equal? variable
					      (lvalue-variable lvalue)))
				     (logical-statement-lvalues statement)))))
			 (if (null? lvalues)
			     #f
			     (make-logical-statement lvalues))))
		       (else #f)))))
		(block-statements block)))))
       (if (null? declarations)
	   (if (and (or (subroutine-block? block)
			(function-block? block))
		    (member variable (bound-variables block)))
	       '()
	       (loop (rest context)))
	   declarations)))))

(define (cotangent-declarations-for variable context)
 (let loop ((context context))
  (if (null? context)
      '()
      (let* ((block (first context))
	     (declarations
	      (removeq
	       #f
	       (map
		(lambda (statement)
		 (if (block? statement)
		     #f
		     (let ((statement (statement-statement statement)))
		      (cond
		       ((external-statement? statement)
			(let ((lvalues
			       (map cotangentify-lvalue
				    (remove-if-not
				     (lambda (lvalue)
				      (equal? variable
					      (lvalue-variable lvalue)))
				     (external-statement-lvalues statement)))))
			 (if (null? lvalues)
			     #f
			     (make-external-statement lvalues))))
		       ((dimension-statement? statement)
			(let ((lvalues
			       (map cotangentify-lvalue
				    (remove-if-not
				     (lambda (lvalue)
				      (equal? variable
					      (lvalue-variable lvalue)))
				     (dimension-statement-lvalues statement)))))
			 (if (null? lvalues)
			     #f
			     (make-dimension-statement lvalues))))
		       ((integer-statement? statement)
			(let ((lvalues
			       (map cotangentify-lvalue
				    (remove-if-not
				     (lambda (lvalue)
				      (equal? variable
					      (lvalue-variable lvalue)))
				     (integer-statement-lvalues statement)))))
			 (if (null? lvalues)
			     #f
			     (make-integer-statement lvalues))))
		       ((real-statement? statement)
			(let ((lvalues
			       (map cotangentify-lvalue
				    (remove-if-not
				     (lambda (lvalue)
				      (equal? variable
					      (lvalue-variable lvalue)))
				     (real-statement-lvalues statement)))))
			 (if (null? lvalues) #f (make-real-statement lvalues))))
		       ((double-precision-statement? statement)
			(let ((lvalues
			       (map cotangentify-lvalue
				    (remove-if-not
				     (lambda (lvalue)
				      (equal? variable
					      (lvalue-variable lvalue)))
				     (double-precision-statement-lvalues
				      statement)))))
			 (if (null? lvalues)
			     #f
			     (make-double-precision-statement lvalues))))
		       ((complex-statement? statement)
			(let ((lvalues
			       (map cotangentify-lvalue
				    (remove-if-not
				     (lambda (lvalue)
				      (equal? variable
					      (lvalue-variable lvalue)))
				     (complex-statement-lvalues statement)))))
			 (if (null? lvalues)
			     #f
			     (make-complex-statement lvalues))))
		       ((logical-statement? statement)
			(let ((lvalues
			       (map cotangentify-lvalue
				    (remove-if-not
				     (lambda (lvalue)
				      (equal? variable
					      (lvalue-variable lvalue)))
				     (logical-statement-lvalues statement)))))
			 (if (null? lvalues)
			     #f
			     (make-logical-statement lvalues))))
		       (else #f)))))
		(block-statements block)))))
       (if (null? declarations)
	   (if (and (or (subroutine-block? block)
			(function-block? block))
		    (member variable (bound-variables block)))
	       '()
	       (loop (rest context)))
	   declarations)))))

(define (declaration? statement)
 (and (not (block? statement))
      (or (external-statement? (statement-statement statement))
	  (dimension-statement? (statement-statement statement))
	  (integer-statement? (statement-statement statement))
	  (real-statement? (statement-statement statement))
	  (double-precision-statement? (statement-statement statement))
	  (complex-statement? (statement-statement statement))
	  (logical-statement? (statement-statement statement)))))

(define (move-declarations-to-beginning statements)
 (append (remove-if-not declaration? statements)
	 (remove-if declaration? statements)))

(define (translate-AD statements reverse-primal?)
 (let loop ((statements statements) (index '()) (context '()))
  (reduce
   append
   (map-indexed
    (lambda (statement k)
     (if (block? statement)
	 (let* ((context (cons statement context))
		(block
		 (make-block
		  (block-beginning statement)
		  (move-declarations-to-beginning
		   (loop (block-statements statement) (cons k index) context))
		  (block-end statement)
		  #f
		  #f
		  #f
		  #f)))
	  (cond
	   ((adf-block? block)
	    (let ((translated-blocks
		   (append
		    (map no-statement-number
			 (map-reduce
			  append
			  '()
			  (lambda (variable)
			   (tangent-declarations-for variable context))
			  (bound-variables block)))
		    (list
		     (make-block
		      (no-statement-number
		       (make-subroutine-statement
			(adf-name index) (bound-variables block)))
		      (append
		       (map no-statement-number
			    (map-reduce append
					'()
					(lambda (variable)
					 (declarations-for variable context))
					(bound-variables block)))
		       (block-statements block))
		      (no-statement-number (make-end-statement #f #f #f))
		      #f
		      #f
		      (adf-independent-variables block)
		      (adf-dependent-variables block)))
		    (translate-adf-opening-assignments block)
		    (list (no-statement-number
			   (make-call-statement
			    (make-forward (adf-name index) #f)
			    (map make-tangent-bundle (bound-variables block)))))
		    (translate-adf-closing-assignments block))))
	     (when *show-AD-translation?*
	      (display "Translating:")
	      (newline)
	      (display-lines (unparse-block block))
	      (display "Translation result:")
	      (newline)
	      (display-lines (unparse-statements translated-blocks))
	      (newline))
	     translated-blocks))
	   ((adr-block? block)
	    (let ((translated-blocks
		   (append
		    (map no-statement-number
			 (map-reduce
			  append
			  '()
			  (lambda (variable)
			   (cotangent-declarations-for variable context))
			  (bound-variables block)))
		    (list
		     (make-block
		      (no-statement-number
		       (make-subroutine-statement
			(adr-name index) (bound-variables block)))
		      (append
		       (map no-statement-number
			    (map-reduce append
					'()
					(lambda (variable)
					 (declarations-for variable context))
					(bound-variables block)))
		       (block-statements block))
		      (no-statement-number (make-end-statement #f #f #f))
		      #f
		      #f
		      (adr-independent-variables block)
		      (adr-dependent-variables block)))
		    ;; This is due to a bug in Tapenade 3.6.
		    (if reverse-primal?
			(list (no-statement-number
			       (make-call-statement
				(adr-name index) (bound-variables block))))
			'())
		    (translate-adr-opening-assignments block)
		    (list
		     (no-statement-number
		      (make-call-statement
		       (make-reverse (adr-name index) #f)
		       (map make-cotangent-bundle (bound-variables block)))))
		    (translate-adr-closing-assignments block))))
	     (when *show-AD-translation?*
	      (display "Translating:")
	      (newline)
	      (display-lines (unparse-block block))
	      (display "Translation result:")
	      (newline)
	      (display-lines (unparse-statements translated-blocks))
	      (newline))
	     translated-blocks))
	   (else (list block))))
	 (list statement)))
    statements)
   '())))

;;; Closure conversion

(define (referenced-subscripts lvalue)
 (cond
  ((unary? lvalue) (internal-error))
  ((binary? lvalue)
   (if (and (or (string? (binary-left lvalue))
		(tangent? (binary-left lvalue))
		(cotangent? (binary-left lvalue)))
	    (parenthesis? (binary-right lvalue))
	    (>= (length (parenthesis-expressions (binary-right lvalue))) 1))
       (map-reduce union
		   '()
		   referenced-variables
		   (parenthesis-expressions (binary-right lvalue)))
       (internal-error)))
  ((parenthesis? lvalue) (internal-error))
  ((forward? lvalue) (internal-error))
  ((tangent? lvalue) '())
  ((tangent-bundle? lvalue) (internal-error))
  ((reverse? lvalue) (internal-error))
  ((cotangent? lvalue) '())
  ((cotangent-bundle? lvalue) (internal-error))
  ((global? lvalue) (internal-error))
  ((string? lvalue) '())
  (else (internal-error))))

(define (referenced-variables statement)
 ;; Only called in new-variables which is only called in lift.
 ;; This can return (co)tangents and (co)tangent bundles but not forwards or
 ;; reverses.
 (cond
  ((statement? statement)
   (referenced-variables (statement-statement statement)))
  ((program-statement? statement) (internal-error))
  ((subroutine-statement? statement) (internal-error))
  ((function-statement? statement) (internal-error))
  ((adf-statement? statement) (internal-error))
  ((adr-statement? statement) (internal-error))
  ((end-statement? statement) '())
  ((external-statement? statement)
   (map-reduce
    union '() referenced-subscripts (external-statement-lvalues statement)))
  ((dimension-statement? statement)
   (map-reduce
    union '() referenced-subscripts (dimension-statement-lvalues statement)))
  ((integer-statement? statement)
   (map-reduce
    union '() referenced-subscripts (integer-statement-lvalues statement)))
  ((real-statement? statement)
   (map-reduce
    union '() referenced-subscripts (real-statement-lvalues statement)))
  ((double-precision-statement? statement)
   (map-reduce union
	       '()
	       referenced-subscripts
	       (double-precision-statement-lvalues statement)))
  ((complex-statement? statement)
   (map-reduce
    union '() referenced-subscripts (complex-statement-lvalues statement)))
  ((logical-statement? statement)
   (map-reduce
    union '() referenced-subscripts (logical-statement-lvalues statement)))
  ((continue-statement? statement) '())
  ((assignment-statement? statement)
   (union (referenced-variables (assignment-statement-lvalue statement))
	  (referenced-variables (assignment-statement-expression statement))))
  ((read-statement? statement)
   (map-reduce
    union '() referenced-variables (read-statement-lvalues statement)))
  ((print-statement? statement)
   (map-reduce
    union '() referenced-variables (print-statement-lvalues statement)))
  ((do-statement? statement)
   (union (referenced-variables (do-statement-assignment-statement statement))
	  (union (referenced-variables (do-statement-expression1 statement))
		 (if (do-statement-expression2 statement)
		     (referenced-variables (do-statement-expression2 statement))
		     '()))))
  ((call-statement? statement)
   (union (referenced-variables (call-statement-name statement))
	  (map-reduce union
		      '()
		      referenced-variables
		      (call-statement-expressions statement))))
  ((block? statement) '())
  ((unary? statement) (referenced-variables (unary-expression statement)))
  ((binary? statement)
   (union (referenced-variables (binary-left statement))
	  (referenced-variables (binary-right statement))))
  ((parenthesis? statement)
   (map-reduce union
	       '()
	       referenced-variables
	       (parenthesis-expressions statement)))
  ;; A forward or reverse can only appear in the target of a call statement and
  ;; always refers to a global (after lifting).
  ((forward? statement) '())
  ((tangent? statement) (list statement))
  ((tangent-bundle? statement) (list statement))
  ((reverse? statement) '())
  ((cotangent? statement) (list statement))
  ((cotangent-bundle? statement) (list statement))
  ((string? statement)
   (if (char-alphabetic? (string-ref statement 0)) (list statement) '()))
  (else (internal-error))))

(define (children block)
 (remove-duplicates
  (map subprogram-name
       (remove-if-not
	(lambda (statement)
	 (or (subroutine-block? statement) (function-block? statement)))
	(block-statements block)))))

(define (declared-variables block)
 (map-reduce union
	     '()
	     (lambda (statement)
	      (remove-duplicates
	       (if (block? statement)
		   '()
		   (let ((statement (statement-statement statement)))
		    (map lvalue-variable
			 (cond ((external-statement? statement)
				(external-statement-lvalues statement))
			       ((dimension-statement? statement)
				(dimension-statement-lvalues statement))
			       ((integer-statement? statement)
				(integer-statement-lvalues statement))
			       ((real-statement? statement)
				(real-statement-lvalues statement))
			       ((double-precision-statement? statement)
				(double-precision-statement-lvalues statement))
			       ((complex-statement? statement)
				(complex-statement-lvalues statement))
			       ((logical-statement? statement)
				(logical-statement-lvalues statement))
			       (else '())))))))
	     (block-statements block)))

(define (free-subscripts lvalue)
 (cond
  ((unary? lvalue) (internal-error))
  ((binary? lvalue)
   (if (and (or (string? (binary-left lvalue))
		(tangent? (binary-left lvalue))
		(cotangent? (binary-left lvalue)))
	    (parenthesis? (binary-right lvalue))
	    (>= (length (parenthesis-expressions (binary-right lvalue))) 1))
       (map-reduce union
		   '()
		   free-variables
		   (parenthesis-expressions (binary-right lvalue)))
       (internal-error)))
  ((parenthesis? lvalue) (internal-error))
  ((forward? lvalue) (internal-error))
  ((tangent? lvalue) '())
  ((tangent-bundle? lvalue) (internal-error))
  ((reverse? lvalue) (internal-error))
  ((cotangent? lvalue) '())
  ((cotangent-bundle? lvalue) (internal-error))
  ((global? lvalue) (internal-error))
  ((string? lvalue) '())
  (else (internal-error))))

(define (free-variables statement)
 ;; Only called in lift.
 (cond
  ((statement? statement) (free-variables (statement-statement statement)))
  ((program-statement? statement) '())
  ((subroutine-statement? statement) '())
  ((function-statement? statement) '())
  ((adf-statement? statement)
   (set-difference
    (map-reduce
     union
     '()
     (lambda (assignment)
      (let loop ((assignment assignment))
       (if (implied-do? assignment)
	   (map-reduce union '() loop (implied-do-assignments assignment))
	   (union
	    (free-variables (assignment-statement-lvalue assignment))
	    (free-variables (assignment-statement-expression assignment))))))
     (adf-statement-assignments statement))
    (bound-variables statement)))
  ((adr-statement? statement)
   (set-difference
    (map-reduce
     union
     '()
     (lambda (assignment)
      (let loop ((assignment assignment))
       (if (implied-do? assignment)
	   (map-reduce union '() loop (implied-do-assignments assignment))
	   (union
	    (free-variables (assignment-statement-lvalue assignment))
	    (free-variables (assignment-statement-expression assignment))))))
     (adr-statement-assignments statement))
    (bound-variables statement)))
  ((end-statement? statement)
   (if (end-statement-assignments statement)
       (set-difference
	(map-reduce
	 union
	 '()
	 (lambda (assignment)
	  (let loop ((assignment assignment))
	   (if (implied-do? assignment)
	       (map-reduce union '() loop (implied-do-assignments assignment))
	       (union
		(free-variables (assignment-statement-lvalue assignment))
		(free-variables
		 (assignment-statement-expression assignment))))))
	 (end-statement-assignments statement))
	(bound-variables statement))
       '()))
  ((external-statement? statement)
   (map-reduce
    union '() free-subscripts (external-statement-lvalues statement)))
  ((dimension-statement? statement)
   (map-reduce
    union '() free-subscripts (dimension-statement-lvalues statement)))
  ((integer-statement? statement)
   (map-reduce union '() free-subscripts (integer-statement-lvalues statement)))
  ((real-statement? statement)
   (map-reduce union '() free-subscripts (real-statement-lvalues statement)))
  ((double-precision-statement? statement)
   (map-reduce
    union '() free-subscripts (double-precision-statement-lvalues statement)))
  ((complex-statement? statement)
   (map-reduce union '() free-subscripts (complex-statement-lvalues statement)))
  ((logical-statement? statement)
   (map-reduce union '() free-subscripts (logical-statement-lvalues statement)))
  ((continue-statement? statement) '())
  ((assignment-statement? statement)
   (union (free-variables (assignment-statement-lvalue statement))
	  (free-variables (assignment-statement-expression statement))))
  ((read-statement? statement)
   (map-reduce union '() free-variables (read-statement-lvalues statement)))
  ((print-statement? statement)
   (map-reduce union '() free-variables (print-statement-lvalues statement)))
  ((do-statement? statement)
   (union (free-variables (do-statement-assignment-statement statement))
	  (union (free-variables (do-statement-expression1 statement))
		 (if (do-statement-expression2 statement)
		     (free-variables (do-statement-expression2 statement))
		     '()))))
  ((call-statement? statement)
   (union (free-variables (call-statement-name statement))
	  (map-reduce union
		      '()
		      free-variables
		      (call-statement-expressions statement))))
  ((block? statement)
   (set-difference
    (union (union (free-variables (block-beginning statement))
		  (free-variables (block-end statement)))
	   (map-reduce union
		       '()
		       free-variables
		       (block-statements statement)))
    (union (union (bound-variables statement) (declared-variables statement))
	   (children statement))))
  ((unary? statement) (free-variables (unary-expression statement)))
  ((binary? statement)
   (union (free-variables (binary-left statement))
	  (free-variables (binary-right statement))))
  ((parenthesis? statement)
   (map-reduce union
	       '()
	       free-variables
	       (parenthesis-expressions statement)))
  ;; A forward or reverse can only appear in the target of a call statement and
  ;; always refers to a global (after lifting).
  ((forward? statement) '())
  ((tangent? statement) (list statement))
  ((tangent-bundle? statement) (list statement))
  ((reverse? statement) '())
  ((cotangent? statement) (list statement))
  ((cotangent-bundle? statement) (list statement))
  ((string? statement)
   (if (char-alphabetic? (string-ref statement 0)) (list statement) '()))
  (else (internal-error))))

(define (new-variables block globals in-scope)
 ;; Only called in lift.
 ;; This can return (co)tangents and (co)tangent bundles but not forwards or
 ;; reverses.
 (union (union (bound-variables block) (declared-variables block))
	(set-difference
	 (map-reduce union '() referenced-variables (block-statements block))
	 (union (children block) (union globals in-scope)))))

(define (rename statement old new)
 ;; Rename the call-site targets and references to a lifted subprogram in the
 ;; body from where it was lifted. Note that we make a scoping design decision
 ;; to do this only in the body from where it was lifted.
 ;; When the block is a function, need to rename all references to the function
 ;; name in the body to its lifted variant as such references are to the
 ;; return value.
 ;; We are lucky that this also renames declarations.
 ;; old is obtained from the name of a lifted subprogram and thus will never be
 ;; a forward, reverse, (co)tangent, or (co)tangent bundle.
 (unless (string? old) (internal-error))
 (let loop ((statement statement))
  (cond
   ((statement? statement)
    (make-statement (statement-number statement)
		    (loop (statement-statement statement))))
   ((program-statement? statement) (internal-error))
   ((subroutine-statement? statement) (internal-error))
   ((function-statement? statement) (internal-error))
   ((adf-statement? statement) (internal-error))
   ((adr-statement? statement) (internal-error))
   ((end-statement? statement) statement)
   ((external-statement? statement)
    (make-external-statement (map loop (external-statement-lvalues statement))))
   ((dimension-statement? statement)
    (make-dimension-statement
     (map loop (dimension-statement-lvalues statement))))
   ((integer-statement? statement)
    (make-integer-statement (map loop (integer-statement-lvalues statement))))
   ((real-statement? statement)
    (make-real-statement (map loop (real-statement-lvalues statement))))
   ((double-precision-statement? statement)
    (make-double-precision-statement
     (map loop (double-precision-statement-lvalues statement))))
   ((complex-statement? statement)
    (make-complex-statement (map loop (complex-statement-lvalues statement))))
   ((logical-statement? statement)
    (make-logical-statement (map loop (logical-statement-lvalues statement))))
   ((continue-statement? statement) statement)
   ((assignment-statement? statement)
    (make-assignment-statement
     (loop (assignment-statement-lvalue statement))
     (loop (assignment-statement-expression statement))))
   ((read-statement? statement)
    (make-read-statement (map loop (read-statement-lvalues statement))))
   ((print-statement? statement)
    (make-print-statement (map loop (print-statement-lvalues statement))))
   ((do-statement? statement)
    (make-do-statement (do-statement-number statement)
		       (loop (do-statement-assignment-statement statement))
		       (loop (do-statement-expression1 statement))
		       (if (do-statement-expression2 statement)
			   (loop (do-statement-expression2 statement))
			   #f)))
   ((call-statement? statement)
    (make-call-statement (loop (call-statement-name statement))
			 (map loop (call-statement-expressions statement))))
   ((block? statement) (internal-error))
   ((unary? statement)
    (make-unary (unary-operator statement) (loop (unary-expression statement))))
   ((binary? statement)
    (make-binary (loop (binary-left statement))
		 (loop (binary-right statement))))
   ((parenthesis? statement)
    (make-parenthesis (map loop (parenthesis-expressions statement))))
   ;; This is called to rename the name of an internal subprogram that is lifted
   ;; when that name appears in a call-site target or a reference in the body.
   ;; The name of the internal subprogram will never be a forward, reverse,
   ;; (co)tangent, or (co)tangent bundle. But the call-site target for f could
   ;; be (forward f) or (reverse f) and when f is lifted to f' the call-site
   ;; target for f is lifted to (forward f') or (reverse f').
   ((forward? statement) (make-forward (loop (forward-name statement)) #f))
   ((tangent? statement) statement)
   ((tangent-bundle? statement) statement)
   ((reverse? statement) (make-reverse (loop (reverse-name statement)) #f))
   ((cotangent? statement) statement)
   ((cotangent-bundle? statement) statement)
   ((string? statement) (if (string=? statement old) new statement))
   (else (internal-error)))))

(define (add-free-variables-to-call-sites statement name free-variables-to-add)
 ;; Add the free variables for a lifted subprogram to each static call site to
 ;; that subprogram in the body from where it was lifted. Note that we make a
 ;; scoping design decision to do this only in the body from where it was
 ;; lifted.
 ;; name is obtained from the names of lifted subprograms and thus will never
 ;; be a forward, reverse, (co)tangent, or (co)tangent bundle.
 (unless (string? name) (internal-error))
 (let loop ((statement statement))
  (cond
   ((statement? statement)
    (make-statement (statement-number statement)
		    (loop (statement-statement statement))))
   ((program-statement? statement) (internal-error))
   ((subroutine-statement? statement) (internal-error))
   ((function-statement? statement) (internal-error))
   ((adf-statement? statement) (internal-error))
   ((adr-statement? statement) (internal-error))
   ((end-statement? statement) statement)
   ((external-statement? statement)
    (make-external-statement (map loop (external-statement-lvalues statement))))
   ((dimension-statement? statement)
    (make-dimension-statement
     (map loop (dimension-statement-lvalues statement))))
   ((integer-statement? statement)
    (make-integer-statement (map loop (integer-statement-lvalues statement))))
   ((real-statement? statement)
    (make-real-statement (map loop (real-statement-lvalues statement))))
   ((double-precision-statement? statement)
    (make-double-precision-statement
     (map loop (double-precision-statement-lvalues statement))))
   ((complex-statement? statement)
    (make-complex-statement (map loop (complex-statement-lvalues statement))))
   ((logical-statement? statement)
    (make-logical-statement (map loop (logical-statement-lvalues statement))))
   ((continue-statement? statement) statement)
   ((assignment-statement? statement)
    (make-assignment-statement
     (loop (assignment-statement-lvalue statement))
     (loop (assignment-statement-expression statement))))
   ((read-statement? statement)
    (make-read-statement (map loop (read-statement-lvalues statement))))
   ((print-statement? statement)
    (make-print-statement (map loop (print-statement-lvalues statement))))
   ((do-statement? statement)
    (make-do-statement (do-statement-number statement)
		       (loop (do-statement-assignment-statement statement))
		       (loop (do-statement-expression1 statement))
		       (if (do-statement-expression2 statement)
			   (loop (do-statement-expression2 statement))
			   #f)))
   ((call-statement? statement)
    ;; ADF and ADR are turned into subroutines, thus a call-site to a
    ;; subroutine might be a forward or reverse. It also can't be a (co)tangent
    ;; or a (co)tangent bundle.
    (if (string=? (cond ((forward? (call-statement-name statement))
			 (forward-name (call-statement-name statement)))
			((reverse? (call-statement-name statement))
			 (reverse-name (call-statement-name statement)))
			(else (call-statement-name statement)))
		  name)
	(make-call-statement
	 (loop (call-statement-name statement))
	 (append (map loop (call-statement-expressions statement))
		 free-variables-to-add))
	(make-call-statement
	 (loop (call-statement-name statement))
	 (map loop (call-statement-expressions statement)))))
   ((block? statement) (internal-error))
   ((unary? statement)
    (make-unary (unary-operator statement) (loop (unary-expression statement))))
   ((binary? statement)
    ;; ADF and ADR are always turned into subroutines not functions, thus a
    ;; call-site to a function cannot be a forward or reverse. It also can't be
    ;; a (co)tangent or a (co)tangent bundle.
    (if (and (string? (binary-left statement))
	     (string=? (binary-left statement) name))
	(cond
	 ((parenthesis? (binary-right statement))
	  (make-binary
	   (loop (binary-left statement))
	   (make-parenthesis
	    (append
	     (map loop (parenthesis-expressions (binary-right statement)))
	     free-variables-to-add))))
	 ((and (binary? (binary-right statement))
	       (parenthesis? (binary-left (binary-right statement))))
	  (make-binary
	   (loop (binary-left statement))
	   (make-binary
	    (make-parenthesis
	     (append (map loop
			  (parenthesis-expressions
			   (binary-left (binary-right statement))))
		     free-variables-to-add))
	    (loop (binary-right (binary-right statement))))))
	 (else (make-binary (loop (binary-left statement))
			    (loop (binary-right statement)))))
	(make-binary (loop (binary-left statement))
		     (loop (binary-right statement)))))
   ((parenthesis? statement)
    (make-parenthesis (map loop (parenthesis-expressions statement))))
   ;; This is called to add the free variables for a lifted subprogram to each
   ;; static call site to that subprogram in the body.
   ;; The name of the internal subprogram will never be a forward, reverse,
   ;; (co)tangent, or (co)tangent bundle.
   ((forward? statement) statement)
   ((tangent? statement) statement)
   ((tangent-bundle? statement) statement)
   ((reverse? statement) statement)
   ((cotangent? statement) statement)
   ((cotangent-bundle? statement) statement)
   ((string? statement) statement)
   (else (internal-error)))))

(define (lift block globals)
 (let loop ((block block)
	    (name #f)
	    (globals globals)
	    (in-scope '())
	    (declarations '())
	    (context (list block)))
  (let* ((name (suffix-name (subprogram-name block) name))
	 ;; Can include (co)tangents and (co)tangent bundles but not forwards
	 ;; or reverses.
	 (the-new-variables (new-variables block globals in-scope))
	 ;; Don't know whether this can include (co)tangents and (co)tangent
	 ;; bundles.
	 (the-free-variables (set-difference (free-variables block)
					     (union globals the-new-variables)))
	 (globals (set-difference globals the-new-variables))
	 (declarations
	  (map (lambda (variable)
		(if (member variable the-new-variables)
		    (map no-statement-number
			 (old-declarations-for variable block))
		    (list-ref declarations (position variable in-scope))))
	       (union in-scope the-new-variables)))
	 ;; Can include (co)tangents and (co)tangent bundles but not forwards or
	 ;; reverses.
	 (in-scope (union in-scope the-new-variables))
	 (lifted-blocks
	  (append
	   ;; Need to decide the scope of the names of internal subprograms.
	   ;; Right now they are not in scope in the bodies of those internal
	   ;; subprograms. If they were, would need to do a rename here as well.
	   ;; And would also need to add the-free-variables to static call sites
	   ;; here as well.
	   (let ((globals (union globals (children block))))
	    (map-reduce
	     append
	     '()
	     (lambda (block)
	      (loop block
		    name
		    globals
		    in-scope
		    declarations
		    (cons block context)))
	     (remove-if-not
	      (lambda (statement)
	       (or (subroutine-block? statement) (function-block? statement)))
	      (block-statements block))))
	   (list
	    (make-block
	     (make-statement
	      (statement-number (block-beginning block))
	      (cond
	       ((program-block? block) (make-program-statement name))
	       ((subroutine-block? block)
		(make-subroutine-statement
		 name
		 (append (subprogram-parameters block) the-free-variables)))
	       ((function-block? block)
		(make-function-statement
		 name
		 (append (subprogram-parameters block)
			 (remove (subprogram-name block) the-free-variables))))
	       (else (internal-error))))
	     (append
	      (map-reduce
	       append
	       '()
	       (lambda (variable)
		(list-ref declarations (position variable in-scope)))
	       (cond ((program-block? block) '())
		     ((subroutine-block? block) the-free-variables)
		     ((function-block? block)
		      (remove (subprogram-name block) the-free-variables))
		     (else (internal-error))))
	      (let ((names (children block))
		    (free-variables-for-each-name
		     (map
		      (lambda (block)
		       (let* ((the-new-variables
			       (new-variables block globals in-scope))
			      (the-free-variables
			       (set-difference
				(free-variables block)
				(union globals the-new-variables))))
			(cond
			 ((subroutine-block? block) the-free-variables)
			 ((function-block? block)
			  (remove (subprogram-name block) the-free-variables))
			 (else (internal-error)))))
		      (remove-if-not (lambda (statement)
				      (or (subroutine-block? statement)
					  (function-block? statement)))
				     (block-statements block)))))
	       (map (lambda (statement)
		     (let loop ((names names)
				(free-variables-for-each-name
				 free-variables-for-each-name)
				(statement statement))
		      (if (null? names)
			  (if (function-block? block)
			      ;; When the block is a function, need to rename
			      ;; all references to the function name in the body
			      ;; to its lifted variant as such references are to
			      ;; the return value.
			      ;; We are lucky that this also renames
			      ;; declarations.
			      (rename statement (subprogram-name block) name)
			      statement)
			  (loop
			   (rest names)
			   (rest free-variables-for-each-name)
			   ;; Rename the call-site targets and references to a
			   ;; lifted subprogram in the body from where it was
			   ;; lifted. Note that we make a scoping design
			   ;; decision to do this only in the body from where it
			   ;; was lifted.
			   ;; We are lucky that this also renames declarations.
			   (rename
			    ;; Add the free variables for a lifted subprogram to
			    ;; each static call site to that subprogram in the
			    ;; body from where it was lifted. Note that we make
			    ;; a scoping design decision to do this only in the
			    ;; body from where it was lifted.
			    (add-free-variables-to-call-sites
			     statement
			     (first names)
			     (first free-variables-for-each-name))
			    (first names)
			    (suffix-name (first names) name))))))
		    (remove-if (lambda (statement)
				(or (subroutine-block? statement)
				    (function-block? statement)))
			       (block-statements block)))))
	     (block-end block)
	     the-free-variables
	     (map (lambda (variable) (declarations-for variable context))
		  the-free-variables)
	     (block-independent-variables block)
	     (block-dependent-variables block))))))
   (when *show-lifting?*
    (display "Lifting:")
    (newline)
    (display-lines (unparse-block block))
    (when #f				;debugging
     (format #t "the-new-variables: ~s~%" the-new-variables)
     (format #t "the-free-variables: ~s~%" the-free-variables)
     (format #t "globals: ~s~%" globals)
     (format #t "in-scope: ~s~%" in-scope))
    (display "Lifting result:")
    (newline)
    (display-lines (unparse-blocks lifted-blocks))
    (newline))
   lifted-blocks)))

;;; Specializer

(define (call-site-with-global-argument blocks)
 ;; When the specializer is called, the program already consists of a set of
 ;; blocks and there are no nested blocks.
 ;; Returns #f if no call-site or a triple (call-site i global) where i is the
 ;; argument index of global.
 (let ((globals (map subprogram-name blocks)))
  (define (is-global? name) (or (global? name) (member name globals)))
  (some
   (lambda (block)
    (some
     (lambda (statement)
      (let loop ((statement statement))
       (cond
	((statement? statement) (loop (statement-statement statement)))
	((program-statement? statement) (internal-error))
	((subroutine-statement? statement) (internal-error))
	((function-statement? statement) (internal-error))
	((adf-statement? statement) (internal-error))
	((adr-statement? statement) (internal-error))
	((end-statement? statement) #f)
	((external-statement? statement)
	 (some loop (external-statement-lvalues statement)))
	((dimension-statement? statement)
	 (some loop (dimension-statement-lvalues statement)))
	((integer-statement? statement)
	 (some loop (integer-statement-lvalues statement)))
	((real-statement? statement)
	 (some loop (real-statement-lvalues statement)))
	((double-precision-statement? statement)
	 (some loop (double-precision-statement-lvalues statement)))
	((complex-statement? statement)
	 (some loop (complex-statement-lvalues statement)))
	((logical-statement? statement)
	 (some loop (logical-statement-lvalues statement)))
	((continue-statement? statement) #f)
	((assignment-statement? statement)
	 (or (loop (assignment-statement-lvalue statement))
	     (loop (assignment-statement-expression statement))))
	((read-statement? statement)
	 (some loop (read-statement-lvalues statement)))
	((print-statement? statement)
	 (some loop (print-statement-lvalues statement)))
	((do-statement? statement)
	 (if (do-statement-expression2 statement)
	     (or (loop (do-statement-assignment-statement statement))
		 (loop (do-statement-expression1 statement))
		 (loop (do-statement-expression2 statement)))
	     (or (loop (do-statement-assignment-statement statement))
		 (loop (do-statement-expression1 statement)))))
	((call-statement? statement)
	 (or (and (some is-global?
			(call-statement-expressions statement))
		  (let ((i (position-if
			    is-global?
			    (call-statement-expressions statement))))
		   (list statement
			 i
			 (list-ref (call-statement-expressions statement) i))))
	     (some loop (call-statement-expressions statement))))
	((block? statement) (internal-error))
	((unary? statement) (loop (unary-expression statement)))
	((binary? statement)
	 (or
	  (and
	   ;; Can never be a forward, reverse, (co)tangent, or (co)tangent
	   ;; bundle.
	   (string? (binary-left statement))
	   (or
	    (and (parenthesis? (binary-right statement))
		 (some is-global?
		       (parenthesis-expressions (binary-right statement)))
		 (let ((i (position-if
			   is-global?
			   (parenthesis-expressions (binary-right statement)))))
		  (list statement
			i
			(list-ref
			 (parenthesis-expressions (binary-right statement))
			 i))))
	    ;; This happens because of the strange way we parse expressions.
	    ;; But it cannot occur deeper.
	    (and (binary? (binary-right statement))
		 (parenthesis? (binary-left (binary-right statement)))
		 (some is-global?
		       (parenthesis-expressions
			(binary-left (binary-right statement))))
		 (let ((i (position-if
			   is-global?
			   (parenthesis-expressions
			    (binary-left (binary-right statement))))))
		  (list statement
			i
			(list-ref (parenthesis-expressions
				   (binary-left (binary-right statement)))
				  i))))))
	  (loop (binary-left statement))
	  (loop (binary-right statement))))
	((parenthesis? statement)
	 (some loop (parenthesis-expressions statement)))
	((forward? statement) #f)
	((tangent? statement) #f)
	((tangent-bundle? statement) #f)
	((reverse? statement) #f)
	((cotangent? statement) #f)
	((cotangent-bundle? statement) #f)
	((string? statement) #f)
	(else (internal-error)))))
     (block-statements block)))
   blocks)))

(define (replace-call-site blocks old new)
 ;; This will replace multiple call sites if they are eq? which is OK.
 (map
  (lambda (block)
   (make-block
    (block-beginning block)
    (map
     (lambda (statement)
      (let loop ((statement statement))
       (cond
	((eq? statement old) new)
	((statement? statement)
	 (make-statement (statement-number statement)
			 (loop (statement-statement statement))))
	((program-statement? statement) (internal-error))
	((subroutine-statement? statement) (internal-error))
	((function-statement? statement) (internal-error))
	((adf-statement? statement) (internal-error))
	((adr-statement? statement) (internal-error))
	((end-statement? statement) statement)
	((external-statement? statement)
	 (make-external-statement
	  (map loop (external-statement-lvalues statement))))
	((dimension-statement? statement)
	 (make-dimension-statement
	  (map loop (dimension-statement-lvalues statement))))
	((integer-statement? statement)
	 (make-integer-statement
	  (map loop (integer-statement-lvalues statement))))
	((real-statement? statement)
	 (make-real-statement (map loop (real-statement-lvalues statement))))
	((double-precision-statement? statement)
	 (make-double-precision-statement
	  (map loop (double-precision-statement-lvalues statement))))
	((complex-statement? statement)
	 (make-complex-statement
	  (map loop (complex-statement-lvalues statement))))
	((logical-statement? statement)
	 (make-logical-statement
	  (map loop (logical-statement-lvalues statement))))
	((continue-statement? statement) statement)
	((assignment-statement? statement)
	 (make-assignment-statement
	  (loop (assignment-statement-lvalue statement))
	  (loop (assignment-statement-expression statement))))
	((read-statement? statement)
	 (make-read-statement (map loop (read-statement-lvalues statement))))
	((print-statement? statement)
	 (make-print-statement (map loop (print-statement-lvalues statement))))
	((do-statement? statement)
	 (make-do-statement (do-statement-number statement)
			    (loop (do-statement-assignment-statement statement))
			    (loop (do-statement-expression1 statement))
			    (if (do-statement-expression2 statement)
				(loop (do-statement-expression2 statement))
				#f)))
	((call-statement? statement)
	 (make-call-statement
	  (loop (call-statement-name statement))
	  (map loop (call-statement-expressions statement))))
	((block? statement) (internal-error))
	((unary? statement)
	 (make-unary (unary-operator statement)
		     (loop (unary-expression statement))))
	((binary? statement)
	 (make-binary (loop (binary-left statement))
		      (loop (binary-right statement))))
	((parenthesis? statement)
	 (make-parenthesis (map loop (parenthesis-expressions statement))))
	((forward? statement) statement)
	((tangent? statement) statement)
	((tangent-bundle? statement) statement)
	((reverse? statement) statement)
	((cotangent? statement) statement)
	((cotangent-bundle? statement) statement)
	((string? statement) statement)
	((global? statement) statement)
	(else (internal-error)))))
     (block-statements block))
    (block-end block)
    (block-closure-variables block)
    (block-declaration-lists block)
    (block-independent-variables block)
    (block-dependent-variables block)))
  blocks))

(define (specialize-target
	 block new-target-name i global closure-variables declaration-lists)
 ;; A target name can never be a forward, reverse, (co)tangent, or (co)tangent
 ;; bundle.
 (unless (string? new-target-name) (internal-error))
 (let ((name (subprogram-name block))
       (parameter (list-ref (subprogram-parameters block) i)))
  ;; parameter is obtained from a subprogram name and thus will never be a
  ;; forward, reverse, (co)tangent, or (co)tangent bundle.
  (unless (string? name) (internal-error))
  ;; parameter is obtained from subprogram parameters and thus will never be
  ;; a forward or reverse.
  ;; It can't be a (co)tangent or (co)tangent bundle.
  (unless (string? parameter) (internal-error))
  (make-block
   (let ((statement (statement-statement (block-beginning block))))
    (no-statement-number
     ((cond ((subroutine-statement? statement) make-subroutine-statement)
	    ((function-statement? statement) make-function-statement)
	    (else (internal-error)))
      new-target-name
      (reduce append
	      (map-indexed
	       (lambda (parameter j)
		(if (= j i)
		    (map (lambda (closure-variable)
			  (closure-variable-name closure-variable parameter))
			 closure-variables)
		    (list parameter)))
	       (subprogram-parameters block))
	      '()))))
   (append
    ;; Add declarations for closure variables.
    (let loop ((parameters (subprogram-parameters block))
	       (j 0)
	       (declarations
		(reduce
		 append
		 (map-indexed
		  (lambda (parameter j)
		   (if (= j i)
		       (map no-statement-number
			    (reduce append declaration-lists '()))
		       '()))
		  (subprogram-parameters block))
		 '())))
     (if (null? parameters)
	 declarations
	 (let ((parameter (first parameters)))
	  (loop (rest parameters)
		(+ j 1)
		(if (= j i)
		    (let loop ((closure-variables closure-variables)
			       (declarations declarations))
		     (if (null? closure-variables)
			 declarations
			 (let ((closure-variable (first closure-variables)))
			  (loop (rest closure-variables)
				(map (lambda (declaration)
				      (rename declaration
					      closure-variable
					      (closure-variable-name
					       closure-variable parameter)))
				     declarations)))))
		    declarations)))))
    ;; Remove empty external declarations.
    (removeq
     #f
     (map
      (lambda (statement)
       (let loop ((statement statement))
	(define (replace-target target)
	 ;; replace parameter that appears as call-site target with global
	 ;; A forward, reverse, (co)tangent, or (co)tangent bundle can never
	 ;; appear in an EXTERNAL (that is grounded by specialization).
	 (if (and (string? target) (string=? target parameter))
	     (cond ((string? global) global)
		   ((global? global) (global-global global))
		   (else (internal-error)))
	     (loop target)))
	(define (replace-argument target arguments)
	 ;; replace parameter that appears as call-site argument with global
	 ;; First need to replace with a global object and let further
	 ;; specialization replace with closure-variables. Otherwise we don't
	 ;; trigger further specialization. We need to replace it with a global
	 ;; object [var foo] instead of the global foo, where var is the
	 ;; original external that was bound to foo, because if the free
	 ;; variables of foo are x, y, and z then ultimately need to rewrite
	 ;; this as var_x, var_y, var_z and not x, y, z.
	 (append
	  (map (lambda (argument)
		;; Forwards, reverses, and (co)tangents can never be arguments.
		(unless (or (string? argument)
			    (tangent-bundle? argument)
			    (cotangent-bundle? argument))
		 (internal-error))
		;; A (co)tangent bundle can never be an external.
		(if (and (string? argument) (string=? argument parameter))
		    (make-global
		     argument
		     ;; By induction, (global-global global) will always be a
		     ;; string and never a global.
		     (cond ((string? global) global)
			   ((global? global) (global-global global))
			   (else (internal-error))))
		    (loop argument)))
	       arguments)
	  ;; A forward, reverse, (co)tangent, or (co)tangent bundle can never
	  ;; appear in an EXTERNAL (that is grounded by specialization).
	  (if (and (string? target) (string=? target parameter))
	      ;; When this becomes a static call site as a result of
	      ;; specialization and the global has closure-variables because it
	      ;; was lifted, need to add them as arguments at the end.
	      (map (lambda (closure-variable)
		    ;; needs work: to add declarations
		    (closure-variable-name closure-variable parameter))
		   closure-variables)
	      '())))
	(cond
	 ((statement? statement)
	  ;; Remove empty external declarations.
	  (let ((statement-statement (loop (statement-statement statement))))
	   (if statement-statement
	       (make-statement (statement-number statement)
			       statement-statement)
	       #f)))
	 ((program-statement? statement) (internal-error))
	 ((subroutine-statement? statement) (internal-error))
	 ((function-statement? statement) (internal-error))
	 ((adf-statement? statement) (internal-error))
	 ((adr-statement? statement) (internal-error))
	 ((end-statement? statement) statement)
	 ((external-statement? statement)
	  ;; Remove external declarations for the specialized parameter. Remove
	  ;; empty external declarations.
	  (let ((variables
		 (map loop
		      (remove parameter
			      (external-statement-lvalues statement)))))
	   (if (null? variables) #f (make-external-statement variables))))
	 ((dimension-statement? statement)
	  (make-dimension-statement
	   (map loop (dimension-statement-lvalues statement))))
	 ((integer-statement? statement)
	  (make-integer-statement
	   (map loop (integer-statement-lvalues statement))))
	 ((real-statement? statement)
	  (make-real-statement
	   (map loop (real-statement-lvalues statement))))
	 ((double-precision-statement? statement)
	  (make-double-precision-statement
	   (map loop (double-precision-statement-lvalues statement))))
	 ((complex-statement? statement)
	  (make-complex-statement
	   (map loop (complex-statement-lvalues statement))))
	 ((logical-statement? statement)
	  (make-logical-statement
	   (map loop (logical-statement-lvalues statement))))
	 ((continue-statement? statement) statement)
	 ((assignment-statement? statement)
	  (make-assignment-statement
	   (loop (assignment-statement-lvalue statement))
	   (loop (assignment-statement-expression statement))))
	 ((read-statement? statement)
	  (make-read-statement (map loop (read-statement-lvalues statement))))
	 ((print-statement? statement)
	  (make-print-statement (map loop (print-statement-lvalues statement))))
	 ((do-statement? statement)
	  (make-do-statement
	   (do-statement-number statement)
	   (loop (do-statement-assignment-statement statement))
	   (loop (do-statement-expression1 statement))
	   (if (do-statement-expression2 statement)
	       (loop (do-statement-expression2 statement))
	       #f)))
	 ((call-statement? statement)
	  (make-call-statement
	   (replace-target (call-statement-name statement))
	   (replace-argument (call-statement-name statement)
			     (call-statement-expressions statement))))
	 ((block? statement) (internal-error))
	 ((unary? statement)
	  (make-unary (unary-operator statement)
		      (loop (unary-expression statement))))
	 ((binary? statement)
	  (cond
	   ((parenthesis? (binary-right statement))
	    (make-binary
	     (replace-target (binary-left statement))
	     (make-parenthesis
	      (replace-argument
	       (binary-left statement)
	       (parenthesis-expressions (binary-right statement))))))
	   ((and (binary? (binary-right statement))
		 (parenthesis? (binary-left (binary-right statement))))
	    (make-binary
	     (replace-target (binary-left statement))
	     (make-binary
	      (make-parenthesis
	       (replace-argument (binary-left statement)
				 (parenthesis-expressions
				  (binary-left (binary-right statement)))))
	      (loop (binary-right (binary-right statement))))))
	   (else (make-binary (loop (binary-left statement))
			      (loop (binary-right statement))))))
	 ((parenthesis? statement)
	  (make-parenthesis (map loop (parenthesis-expressions statement))))
	 ((forward? statement) statement)
	 ((tangent? statement) statement)
	 ((tangent-bundle? statement) statement)
	 ((reverse? statement) statement)
	 ((cotangent? statement) statement)
	 ((cotangent-bundle? statement) statement)
	 ;; Rename the return value for a function (it can never be a forward,
	 ;; reverse, (co)tangent, or (co)tangent bundle).
	 ;; We are lucky that this also renames declarations.
	 ((string? statement)
	  (if (string=? statement name) new-target-name statement))
	 (else (internal-error)))))
      (block-statements block))))
   (block-end block)
   (block-closure-variables block)
   (block-declaration-lists block)
   (block-independent-variables block)
   (block-dependent-variables block))))

(define (call-site-name call-site)
 (cond ((call-statement? call-site) (call-statement-name call-site))
       ((binary? call-site) (binary-left call-site))
       (else (internal-error))))

(define (call-site-arguments call-site)
 (cond ((call-statement? call-site) (call-statement-expressions call-site))
       ((binary? call-site)
	(cond ((parenthesis? (binary-right call-site))
	       (parenthesis-expressions (binary-right call-site)))
	      ((and (binary? (binary-right call-site))
		    (parenthesis? (binary-left (binary-right call-site))))
	       (parenthesis-expressions (binary-left (binary-right call-site))))
	      (else (internal-error))))
       (else (internal-error))))

(define (new-call-site call-site new-call-site-name i closure-variables)
 (let ((new-arguments
	(reduce
	 append
	 (map-indexed
	  (lambda (argument j)
	   (if (= j i)
	       (cond ((string? argument) closure-variables)
		     ((global? argument)
		      (map (lambda (closure-variable)
			    (closure-variable-name closure-variable
						   (global-name argument)))
			   closure-variables))
		     (else (internal-error)))
	       (list argument)))
	  (call-site-arguments call-site))
	 '())))
  (cond
   ((call-statement? call-site)
    (make-call-statement new-call-site-name new-arguments))
   ((binary? call-site)
    (cond ((parenthesis? (binary-right call-site))
	   (make-binary new-call-site-name (make-parenthesis new-arguments)))
	  ((and (binary? (binary-right call-site))
		(parenthesis? (binary-left (binary-right call-site))))
	   (make-binary new-call-site-name
			(make-binary (make-parenthesis new-arguments)
				     (binary-right (binary-right call-site)))))
	  (else (internal-error))))
   (else (internal-error)))))

(define (specialize blocks)
 (let loop ((blocks blocks) (n 0))
  (let ((result (call-site-with-global-argument blocks)))
   (if result
       (let* ((call-site (first result))
	      (i (second result))
	      (global (third result))
	      (name (call-site-name call-site))
	      ;; When specializing a call to (forward f) or (reverse f) need
	      ;; to specialize f.
	      ;; The target is renamed to (specialize f n).
	      (new-target-name
	       (specialize-name (cond ((forward? name) (forward-name name))
				      ((reverse? name) (reverse-name name))
				      (else name))
				n))
	      ;; But the call size is renamed to (forward (specialize f n))
	      ;; or (reverse (specialize f n)).
	      (new-call-site-name
	       (cond ((forward? name)
		      (make-forward (specialize-name (forward-name name) n) #f))
		     ((reverse? name)
		      (make-reverse (specialize-name (reverse-name name) n) #f))
		     (else (specialize-name name n))))
	      (target
	       ;; (subprogram-name block) will never be a forward or reverse.
	       ;; When specializing a call to (forward f) or (reverse f) need
	       ;; to specialize f.
	       (find-if (lambda (block)
			 (string=? (subprogram-name block)
				   (cond ((forward? name) (forward-name name))
					 ((reverse? name) (reverse-name name))
					 (else name))))
			blocks))
	      (other-block
	       (find-if
		(lambda (block)
		 ;; global will never be a forward or reverse.
		 (string=? (subprogram-name block)
			   (cond ((string? global) global)
				 ((global? global) (global-global global))
				 (else (internal-error)))))
		blocks))
	      (closure-variables (block-closure-variables other-block))
	      (declaration-lists (block-declaration-lists other-block))
	      (specialized-call-site
	       (new-call-site
		call-site new-call-site-name i closure-variables))
	      (specialized-target (specialize-target target
						     new-target-name
						     i
						     global
						     closure-variables
						     declaration-lists))
	      (specialized-blocks
	       (cons
		specialized-target
		;; needs work: to add/rename declarations
		(replace-call-site blocks call-site specialized-call-site))))
	(when *show-specialization?*
	 (display "Specializing call site:")
	 (newline)
	 (display-lines (unparse-call-site call-site))
	 (format #t "i: ~s~%" i)
	 (display "Specialized call site:")
	 (newline)
	 (display-lines (unparse-call-site specialized-call-site))
	 (display "Unspecialized target:")
	 (newline)
	 (display-lines (unparse-block target))
	 (newline)
	 (display "Specialized target:")
	 (newline)
	 (display-lines (unparse-block specialized-target))
	 (newline)
	 (display "Specialized program:")
	 (newline)
	 (display-lines (unparse-blocks specialized-blocks))
	 (newline))
	(loop specialized-blocks (+ n 1)))
       blocks))))

;;; Tree shaker

(define (call-site-to? name blocks)
 ;; When the tree shaker is called, the program already consists of a set of
 ;; blocks and there are no nested blocks.
 ;; name is obtained from the names of lifted subprograms and thus will never
 ;; be a forward, reverse, (co)tangent, or (co)tangent bundle.
 (unless (string? name) (internal-error))
 (some
  (lambda (block)
   (some
    (lambda (statement)
     (let loop ((statement statement))
      (cond
       ((statement? statement) (loop (statement-statement statement)))
       ((program-statement? statement) (internal-error))
       ((subroutine-statement? statement) (internal-error))
       ((function-statement? statement) (internal-error))
       ((adf-statement? statement) (internal-error))
       ((adr-statement? statement) (internal-error))
       ((end-statement? statement) #f)
       ((external-statement? statement)
	(some loop (external-statement-lvalues statement)))
       ((dimension-statement? statement)
	(some loop (dimension-statement-lvalues statement)))
       ((integer-statement? statement)
	(some loop (integer-statement-lvalues statement)))
       ((real-statement? statement)
	(some loop (real-statement-lvalues statement)))
       ((double-precision-statement? statement)
	(some loop (double-precision-statement-lvalues statement)))
       ((complex-statement? statement)
	(some loop (complex-statement-lvalues statement)))
       ((logical-statement? statement)
	(some loop (logical-statement-lvalues statement)))
       ((continue-statement? statement) #f)
       ((assignment-statement? statement)
	(or (loop (assignment-statement-lvalue statement))
	    (loop (assignment-statement-expression statement))))
       ((read-statement? statement)
	(some loop (read-statement-lvalues statement)))
       ((print-statement? statement)
	(some loop (print-statement-lvalues statement)))
       ((do-statement? statement)
	(if (do-statement-expression2 statement)
	    (or (loop (do-statement-assignment-statement statement))
		(loop (do-statement-expression1 statement))
		(loop (do-statement-expression2 statement)))
	    (or (loop (do-statement-assignment-statement statement))
		(loop (do-statement-expression1 statement)))))
       ((call-statement? statement)
	(or (and (string? (call-statement-name statement))
		 (string=? (call-statement-name statement) name))
	    ;; A subroutine call site to (forward f) or (reverse f) counts as
	    ;; a reference to f.
	    (and (forward? (call-statement-name statement))
		 (string=? (forward-name (call-statement-name statement)) name))
	    (and (reverse? (call-statement-name statement))
		 (string=? (reverse-name (call-statement-name statement)) name))
	    (some loop (call-statement-expressions statement))))
       ((block? statement) (internal-error))
       ((unary? statement) (loop (unary-expression statement)))
       ((binary? statement)
	(or
	 (and
	  ;; A function call site cannot be to a forward, reverse, (co)tangent,
	  ;; or (co)tangent bundle.
	  (string? (binary-left statement))
	  (string=? (binary-left statement) name)
	  (or (parenthesis? (binary-right statement))
	      ;; This happens because of the strange way we parse expressions.
	      ;; But it cannot occur deeper.
	      (and (binary? (binary-right statement))
		   (parenthesis? (binary-left (binary-right statement))))))
	 (loop (binary-left statement))
	 (loop (binary-right statement))))
       ((parenthesis? statement)
	(some loop (parenthesis-expressions statement)))
       ((forward? statement) #f)
       ((tangent? statement) #f)
       ((tangent-bundle? statement) #f)
       ((reverse? statement) #f)
       ((cotangent? statement) #f)
       ((cotangent-bundle? statement) #f)
       ((string? statement) #f)
       (else (internal-error)))))
    (block-statements block)))
  blocks))

(define (tree-shake root blocks)
 ;; This works because there is no recursion in Fortran.
 (let loop ((blocks blocks))
  (let ((unused-blocks
	 (remove-if (lambda (block)
		     (or (string=? (subprogram-name block) root)
			 (call-site-to? (subprogram-name block) blocks)))
		    blocks)))
   (cond ((null? unused-blocks) blocks)
	 (else
	  (when *show-tree-shake?*
	   (display "Tree shake:")
	   (newline)
	   (display-lines (unparse-blocks unused-blocks))
	   (newline))
	  (loop (set-differenceq blocks unused-blocks)))))))

;;; AD

(define (targets block)
 ;; When the tree shaker is called, the program already consists of a set of
 ;; blocks and there are no nested blocks.
 (map-reduce
  union
  '()
  (lambda (statement)
   (let loop ((statement statement))
    (cond
     ((statement? statement) (loop (statement-statement statement)))
     ((program-statement? statement) (internal-error))
     ((subroutine-statement? statement) (internal-error))
     ((function-statement? statement) (internal-error))
     ((adf-statement? statement) (internal-error))
     ((adr-statement? statement) (internal-error))
     ((end-statement? statement) '())
     ((external-statement? statement)
      (map-reduce union '() loop (external-statement-lvalues statement)))
     ((dimension-statement? statement)
      (map-reduce union '() loop (dimension-statement-lvalues statement)))
     ((integer-statement? statement)
      (map-reduce union '() loop (integer-statement-lvalues statement)))
     ((real-statement? statement)
      (map-reduce union '() loop (real-statement-lvalues statement)))
     ((double-precision-statement? statement)
      (map-reduce
       union '() loop (double-precision-statement-lvalues statement)))
     ((complex-statement? statement)
      (map-reduce union '() loop (complex-statement-lvalues statement)))
     ((logical-statement? statement)
      (map-reduce union '() loop (logical-statement-lvalues statement)))
     ((continue-statement? statement) '())
     ((assignment-statement? statement)
      (union (loop (assignment-statement-lvalue statement))
	     (loop (assignment-statement-expression statement))))
     ((read-statement? statement)
      (map-reduce union '() loop (read-statement-lvalues statement)))
     ((print-statement? statement)
      (map-reduce union '() loop (print-statement-lvalues statement)))
     ((do-statement? statement)
      (union (loop (do-statement-assignment-statement statement))
	     (union (loop (do-statement-expression1 statement))
		    (if (do-statement-expression2 statement)
			(loop (do-statement-expression2 statement))
			'()))))
     ((call-statement? statement)
      (adjoin
       (call-statement-name statement)
       (map-reduce union '() loop (call-statement-expressions statement))))
     ((block? statement) (internal-error))
     ((unary? statement) (loop (unary-expression statement)))
     ((binary? statement)
      (let ((tail (union (loop (binary-left statement))
			 (loop (binary-right statement)))))
       (if (or (parenthesis? (binary-right statement))
	       ;; This happens because of the strange way we parse expressions.
	       ;; But it cannot occur deeper.
	       (and (binary? (binary-right statement))
		    (parenthesis? (binary-left (binary-right statement)))))
	   (adjoin (binary-left statement) tail)
	   tail)))
     ((parenthesis? statement)
      (map-reduce union '() loop (parenthesis-expressions statement)))
     ((forward? statement) '())
     ((tangent? statement) '())
     ((tangent-bundle? statement) '())
     ((reverse? statement) '())
     ((cotangent? statement) '())
     ((cotangent-bundle? statement) '())
     ((string? statement) '())
     (else (internal-error)))))
  (block-statements block)))

(define (deforwardreverseize name)
 (cond ((forward? name) (forward-name name))
       ((reverse? name) (reverse-name name))
       (else name)))

(define (call-graph blocks)
 (let* ((vertices (map subprogram-name blocks))
	(edges
	 (map-reduce
	  append
	  '()
	  (lambda (block)
	   (map (lambda (target) (list (subprogram-name block) target))
		(remove-if-not
		 (lambda (target)
		  (member (deforwardreverseize target) vertices))
		 (targets block))))
	  blocks)))
  (list vertices edges)))

(define (transitive-closure graph)
 (let ((vertices (first graph)))
  (let loop ((edges (second graph)))
   (let ((new-edges
	  (set-difference
	   (map-reduce
	    union
	    '()
	    (lambda (edge1)
	     (map-reduce union
			 '()
			 (lambda (edge2)
			  (if (equal? (deforwardreverseize (second edge1))
				      (deforwardreverseize (first edge2)))
			      (list (list (first edge1) (second edge2)))
			      '()))
			 edges))
	    edges)
	   edges)))
    (if (null? new-edges)
	(list vertices edges)
	(loop (union edges new-edges)))))))

(define (subprograms-needing-AD blocks)
 (let* ((graph (transitive-closure (call-graph blocks)))
	(edges (second graph))
	(AD-targets
	 (remove-duplicates
	  (remove-if-not
	   (lambda (vertex) (or (forward? vertex) (reverse? vertex)))
	   (map second edges)))))
  (topological-sort
   (lambda (x y) (member (deforwardreverseize (first x)) (second y)))
   (map
    (lambda (AD-target)
     (list AD-target
	   (map (lambda (edge) (deforwardreverseize (second edge)))
		(remove-if-not (lambda (edge)
				(and (string=? (first edge)
					       (deforwardreverseize AD-target))
				     (or (forward? (second edge))
					 (reverse? (second edge)))))
			       edges))))
    AD-targets))))

;;; Unparser

(define (unparse-expression expression)
 (cond ((unary? expression)
	(string-append (unary-operator expression)
		       (unparse-expression (unary-expression expression))))
       ((binary? expression)
	(string-append (unparse-expression (binary-left expression))
		       (unparse-expression (binary-right expression))))
       ((parenthesis? expression)
	(string-append
	 "("
	 (commas-between
	  (map unparse-expression (parenthesis-expressions expression)))
	 ")"))
       ((forward? expression) (forwardize-name expression))
       ((tangent? expression)
	(tangent-variable-name (unparse-expression (tangent-name expression))))
       ((tangent-bundle? expression)
	(string-append
	 (unparse-expression (tangent-bundle-name expression))
	 ", "
	 (tangent-variable-name
	  (unparse-expression (tangent-bundle-name expression)))))
       ((reverse? expression) (reverseize-name expression))
       ((cotangent? expression)
	(cotangent-variable-name
	 (unparse-expression (cotangent-name expression))))
       ((cotangent-bundle? expression)
	(string-append
	 (unparse-expression (cotangent-bundle-name expression))
	 ", "
	 (cotangent-variable-name
	  (unparse-expression (cotangent-bundle-name expression)))))
       ((global? expression)
	;; This is not legitimate Fortran and is just to unparse the
	;; intermediate states of specialization.
	(string-append "["
		       (unparse-expression (global-name expression))
		       ", "
		       (unparse-expression (global-global expression))
		       "]"))
       ((string? expression) expression)
       (else (internal-error))))

;;; The next two are because I am lazy.
(define (unparse-variable variable) (unparse-expression variable))

(define (unparse-lvalue lvalue) (unparse-expression lvalue))

(define (unparse-assignment statement)
 (string-append
  (unparse-lvalue (assignment-statement-lvalue statement))
  " = "
  (unparse-expression (assignment-statement-expression statement))))

(define (unparse-do-statement statement)
 (string-append
  (unparse-assignment (do-statement-assignment-statement statement))
  ", "
  (unparse-expression (do-statement-expression1 statement))
  (if (do-statement-expression2 statement)
      (string-append
       ", " (unparse-expression (do-statement-expression2 statement)))
      "")))

(define (unparse-optional-implied-do statement)
 (if (implied-do? statement)
     (string-append
      "("
      (commas-between
       (append
	(map unparse-optional-implied-do
	     (implied-do-assignments statement))
	(list (unparse-do-statement (implied-do-do-statement statement)))))
      ")")
     (unparse-assignment statement)))

(define (break-into-continuation-lines line)
 (if (<= (string-length line) 72)
     (list line)
     (let loop ((line (substring line 72 (string-length line)))
		(lines (list (substring line 0 72))))
      (if (> (string-length line) (- 72 6))
	  (loop (substring line (- 72 6) (string-length line))
		(cons (string-append "     +" (substring line 0 (- 72 6)))
		      lines))
	  (reverse (cons (string-append "     +" line) lines))))))

(define (unparse-statement statement)
 (break-into-continuation-lines
  (string-append
   " "
   (statement-number statement)
   " "
   (let ((statement (statement-statement statement)))
    (cond
     ((program-statement? statement)
      (string-append "program " (program-statement-name statement)))
     ((subroutine-statement? statement)
      (string-append "subroutine "
		     (subroutine-statement-name statement)
		     "("
		     (commas-between (subroutine-statement-variables statement))
		     ")"))
     ((function-statement? statement)
      (string-append "function "
		     (function-statement-name statement)
		     "("
		     (commas-between (function-statement-variables statement))
		     ")"))
     ((adf-statement? statement)
      (string-append
       "adf("
       (commas-between
	(map unparse-optional-implied-do (adf-statement-assignments statement)))
       ")"))
     ((adr-statement? statement)
      (string-append
       "adr("
       (commas-between
	(map unparse-optional-implied-do (adr-statement-assignments statement)))
       ")"))
     ((end-statement? statement)
      (case (end-statement-kind statement)
       ((program)
	(if (end-statement-name statement)
	    (string-append "end program " (end-statement-name statement))
	    "end program"))
       ((subroutine)
	(if (end-statement-name statement)
	    (string-append "end subroutine " (end-statement-name statement))
	    "end subroutine"))
       ((function)
	(if (end-statement-name statement)
	    (string-append "end function " (end-statement-name statement))
	    "end function"))
       ((do) "end do")
       ((adf)
	(string-append "end adf("
		       (commas-between
			(map unparse-optional-implied-do
			     (end-statement-assignments statement)))
		       ")"))
       ((adr)
	(string-append "end adr("
		       (commas-between
			(map unparse-optional-implied-do
			     (end-statement-assignments statement)))
		       ")"))
       (else
	(cond ((end-statement-name statement)
	       (string-append "end " (end-statement-name statement)))
	      ;; We don't handle the case of an unnamed end statement with
	      ;; closing assignments anymore because we can't determine whether
	      ;; to look for tangent or cotangent closing assignments.
	      (else "end")))))
     ((external-statement? statement)
      (string-append
       "external "
       (commas-between
	(map unparse-lvalue (external-statement-lvalues statement)))))
     ((dimension-statement? statement)
      (string-append
       "dimension "
       (commas-between
	(map unparse-lvalue (dimension-statement-lvalues statement)))))
     ((integer-statement? statement)
      (string-append
       "integer "
       (commas-between
	(map unparse-lvalue (integer-statement-lvalues statement)))))
     ((real-statement? statement)
      (string-append
       "real "
       (commas-between
	(map unparse-lvalue (real-statement-lvalues statement)))))
     ((double-precision-statement? statement)
      (string-append
       "double precision "
       (commas-between
	(map unparse-lvalue (double-precision-statement-lvalues statement)))))
     ((complex-statement? statement)
      (string-append
       "complex "
       (commas-between
	(map unparse-lvalue (complex-statement-lvalues statement)))))
     ((logical-statement? statement)
      (string-append
       "logical "
       (commas-between
	(map unparse-lvalue (logical-statement-lvalues statement)))))
     ((continue-statement? statement) "continue")
     ((assignment-statement? statement) (unparse-assignment statement))
     ((read-statement? statement)
      (string-append
       "read *, "
       (commas-between
	(map unparse-lvalue (read-statement-lvalues statement)))))
     ((print-statement? statement)
      (string-append
       "print *, "
       (commas-between
	(map unparse-lvalue (print-statement-lvalues statement)))))
     ((do-statement? statement)
      (string-append "do "
		     (if (do-statement-number statement)
			 (string-append (do-statement-number statement) " ")
			 "")
		     (unparse-do-statement statement)))
     ((call-statement? statement)
      (string-append
       "call "
       (unparse-variable (call-statement-name statement))
       "("
       (commas-between
	(map unparse-expression (call-statement-expressions statement)))
       ")"))
     (else (internal-error)))))))

(define (unparse-call-site call-site)
 (cond ((binary? call-site) (list (unparse-expression call-site)))
       ((call-statement? call-site)
	(unparse-statement (no-statement-number call-site)))
       (else (internal-error))))

(define (unparse-statements statements)
 (map-reduce append
	     '()
	     (lambda (statement)
	      (if (block? statement)
		  (append (unparse-statement (block-beginning statement))
			  (unparse-statements (block-statements statement))
			  (unparse-statement (block-end statement)))
		  (unparse-statement statement)))
	     statements))

(define (unparse-block block)
 (unless (block? block) (internal-error))
 (append (unparse-statement (block-beginning block))
	 (unparse-statements (block-statements block))
	 (unparse-statement (block-end block))))

(define (unparse-blocks blocks)
 (reduce (lambda (lines1 lines2) (append lines1 '("") lines2))
	 (map unparse-block blocks)
	 '()))

(define (display-line line) (display line) (newline))

(define (display-lines lines) (for-each display-line lines))

;;; Commands

;;; Top Level

(define-command (main (at-most-one
		       ("tapenade-path" tapenade-path?
			(tapenade-path "tapenade-path" string-argument "")))
		      (at-most-one ("show-AD-translation" show-AD-translation?))
		      (at-most-one ("show-lifting" show-lifting?))
		      (at-most-one ("show-specialization" show-specialization?))
		      (at-most-one ("show-tree-shake" show-tree-shake?))
		      (at-most-one ("show-commands" show-commands?))
		      (at-most-one ("show-all" show-all?))
		      (at-most-one ("no-lifting" no-lifting?))
		      (at-most-one ("no-tree-shake" no-tree-shake?))
		      ;; The default is tapenade.
		      (at-most-one ("adifor" adifor?)
				   ("tapenade" tapenade?)
				   ("no-AD" no-AD?))
		      (at-most-one ("no-Fortran-compiler" no-Fortran-compiler?))
		      (at-most-one ("keep" keep?))
		      (at-most-one ("reverse-primal" reverse-primal?))
		      (at-most-one
		       ("fc" fc?
			(fc "Fortran compiler" string-argument "gfortran")))
		      (required (pathname "pathname" string-argument)))
 (unless tapenade-path?
  (set! tapenade-path
	(if (getenv "FARFALLEN_TAPENADE_PATH")
	    (getenv "FARFALLEN_TAPENADE_PATH")
	    ;; /home/barak/src/git/adic/tapenade3.6
	    "/home/qobi/pkg/tapenade3.6")))
 (set! *adifor?* adifor?)
 (set! *show-AD-translation?* (or show-AD-translation? show-all?))
 (set! *show-lifting?* (or show-lifting? show-all?))
 (set! *show-specialization?* (or show-specialization? show-all?))
 (set! *show-tree-shake?* (or show-tree-shake? show-all?))
 (let ((blocks
	(translate-AD
	 (get-blocks
	  (process-lines (read-file (default-extension pathname "farfel"))))
	 reverse-primal?)))
  (if no-lifting?
      (write-file (unparse-blocks blocks) (replace-extension pathname "f"))
      (let* ((globals (remove-duplicates (map subprogram-name blocks)))
	     (blocks (specialize
		      (map-reduce append
				  '()
				  (lambda (block) (lift block globals))
				  blocks))))
       (if no-tree-shake?
	   (write-file (unparse-blocks blocks) (replace-extension pathname "f"))
	   (let ((blocks (tree-shake
			  (subprogram-name (find-if program-block? blocks))
			  blocks)))
	    (define (adifor)
	     (let ((subprograms-needing-AD (subprograms-needing-AD blocks)))
	      (when (some (lambda (subprogram) (reverse? (first subprogram)))
			  subprograms-needing-AD)
	       (panic "Can't use ADR with -adifor"))
	      (for-each-indexed (lambda (subprogram i)
				 (set-forward-epsilon! (first subprogram) i))
				subprograms-needing-AD)
	      (write-file (unparse-blocks blocks)
			  (replace-extension pathname "f"))
	      (let loop ((remaining-subprograms-needing-AD
			  subprograms-needing-AD)
			 (i 0)
			 (previous '()))
	       (if (null? remaining-subprograms-needing-AD)
		   (unless no-Fortran-compiler?
		    (let ((command
			   (format #f
				   "~a -o ~a ~a"
				   fc
				   (strip-extension pathname)
				   (spaces-between
				    (cons (replace-extension pathname "f")
					  previous)))))
		     (when (or show-commands? show-all?)
		      (display command)
		      (newline))
		     (system command))
		    (unless keep?
		     (rm "AD_cache")
		     (rm (replace-extension pathname "f"))
		     (for-each (lambda (pathname)
				(rm (replace-extension pathname "f"))
				;; needs work: breaks if there are directories
				(rm (string-append
				     "." (replace-extension pathname "f")))
				(rm (replace-extension pathname "aux"))
				(rm (replace-extension pathname "A")))
			       previous)
		     (for-each-n
		      (lambda (i)
		       (rm (replace-extension
			    (string-append
			     (adifor-prefix i) "_" (strip-extension pathname))
			    "f"))
		       ;; needs work: breaks if there are directories
		       (rm (string-append
			    "."
			    (replace-extension
			     (string-append
			      (adifor-prefix i) "_" (strip-extension pathname))
			     "f")))
		       (rm (replace-extension
			    (string-append
			     (adifor-prefix i) "_" (strip-extension pathname))
			    "adf"))
		       (rm (replace-extension
			    (string-append
			     (adifor-prefix i) "_" (strip-extension pathname))
			    "cmp")))
		      (length subprograms-needing-AD))))
		   (let* ((subprogram (first remaining-subprograms-needing-AD))
			  (block
			   (find-if
			    (lambda (block)
			     (string=? (subprogram-name block)
				       (forward-name (first subprogram))))
			    blocks))
			  (command
			   (format #f "Adifor2.1 AD_SCRIPT=~a"
				   (replace-extension
				    (string-append (adifor-prefix i)
						   "_"
						   (strip-extension pathname))
				    "adf"))))
		    (write-file
		     (unparse-blocks
		      (tree-shake (forward-name (first subprogram)) blocks))
		     (replace-extension
		      (string-append
		       (adifor-prefix i) "_" (strip-extension pathname))
		      "f"))
		    (write-file
		     (list
		      (format #f "AD_TOP               = ~a"
			      (forward-name (first subprogram)))
		      (format #f "AD_IVARS             = ~a"
			      (commas-between
			       (block-independent-variables block)))
		      (format #f "AD_DVARS             = ~a"
			      (commas-between
			       (block-dependent-variables block)))
		      (format #f "AD_PROG              = ~a"
			      (replace-extension
			       (string-append (adifor-prefix i)
					      "_"
					      (strip-extension pathname))
			       "cmp"))
		      (format #f "AD_PREFIX            = ~a" (adifor-prefix i))
		      "AD_PMAX              = 1"
		      "AD_OUTPUT_DIR        = ."
		      "AD_SUPPRESS_LDG      = true"
		      "AD_SUPPRESS_NUM_COLS = true"
		      "AD_EXCEPTION_FLAVOR  = performance"
		      "AD_SCALAR_GRADIENTS  = true")
		     (replace-extension
		      (string-append
		       (adifor-prefix i) "_" (strip-extension pathname))
		      "adf"))
		    (write-file
		     (cons (replace-extension
			    (string-append (adifor-prefix i)
					   "_"
					   (strip-extension pathname))
			    "f")
			   previous)
		     (replace-extension
		      (string-append
		       (adifor-prefix i) "_" (strip-extension pathname))
		      "cmp"))
		    (when (or show-commands? show-all?)
		     (display command)
		     (newline))
		    (system command)
		    (loop
		     (rest remaining-subprograms-needing-AD)
		     (+ i 1)
		     (append
		      previous
		      (cons (replace-extension
			     (string-append (adifor-prefix i)
					    "_"
					    (adifor-prefix i)
					    "_"
					    (strip-extension pathname))
			     "f")
			    (remove-if-not
			     file-exists?
			     (map (lambda (pathname)
				   (replace-extension
				    (string-append (adifor-prefix i)
						   "_"
						   (strip-extension pathname))
				    "f"))
				  previous))))))))))
	    (define (tapenade)
	     (let ((subprograms-needing-AD (subprograms-needing-AD blocks)))
	      (for-each-indexed (lambda (subprogram i)
				 ((cond ((forward? (first subprogram))
					 set-forward-epsilon!)
					((reverse? (first subprogram))
					 set-reverse-epsilon!)
					(else (internal-error)))
				  (first subprogram) i))
				subprograms-needing-AD)
	      (write-file (unparse-blocks blocks)
			  (replace-extension pathname "f"))
	      (for-each-indexed
	       (lambda (subprogram i)
		(let* ((mode (cond ((forward? (first subprogram)) 'forward)
				   ((reverse? (first subprogram)) 'reverse)
				   (else (internal-error))))
		       (block
			(find-if
			 (lambda (block)
			  (string=? (subprogram-name block)
				    (deforwardreverseize (first subprogram))))
			 blocks))
		       (command
			(format
			 #f
			 "tapenade3.6 -root ~a -vars \"~a\" -outvars \"~a\" ~a -o ~a -diffvarname _~a -difffuncname _~a ~a"
			 (deforwardreverseize (first subprogram))
			 ;; 3.6 appears to require commas, not spaces
			 ;; 2.2.2 appears not to accept either commas or spaces
			 ;; 3.6 complains if outvars are not included in vars
			 ;; in reverse mode
			 ;; 2.2.2 does not complain
			 (commas-without-spaces-between
			  (case mode
			   ((forward) (block-independent-variables block))
			   ((reverse)
			    (union (block-independent-variables block)
				   (block-dependent-variables block)))
			   (else (internal-error))))
			 (commas-without-spaces-between
			  (block-dependent-variables block))
			 (case mode
			  ((forward) "-d")
			  ((reverse) "-b")
			  (else (internal-error)))
			 (strip-extension pathname)
			 ((case mode
			   ((forward) tapenade-forward-suffix)
			   ((reverse) tapenade-reverse-suffix)
			   (else (internal-error)))
			  i)
			 ((case mode
			   ((forward) tapenade-forward-suffix)
			   ((reverse) tapenade-reverse-suffix)
			   (else (internal-error)))
			  i)
			 (spaces-between
			  (cons
			   (replace-extension pathname "f")
			   (map-n (lambda (j)
				   (replace-extension
				    (string-append
				     (strip-extension pathname)
				     "_"
				     ((cond
				       ((forward?
					 (first
					  (list-ref subprograms-needing-AD j)))
					tapenade-forward-suffix)
				       ((reverse?
					 (first
					  (list-ref subprograms-needing-AD j)))
					tapenade-reverse-suffix)
				       (else (internal-error)))
				      j))
				    "f"))
				  i))))))
		 (when (or show-commands? show-all?)
		  (display command)
		  (newline))
		 (system command)))
	       subprograms-needing-AD)
	      (unless no-Fortran-compiler?
	       (let ((command
		      (format #f
			      "~a -o ~a ~a ~a/ADFirstAidKit/adBuffer.f ~a/ADFirstAidKit/adStack.c"
			      fc
			      (strip-extension pathname)
			      (spaces-between
			       (cons
				(replace-extension pathname "f")
				(map-indexed
				 (lambda (subprogram i)
				  (let ((mode
					  (cond
					   ((forward? (first subprogram))
					    'forward)
					   ((reverse? (first subprogram))
					    'reverse)
					   (else
					    (internal-error)))))
				   (replace-extension
				    (string-append
				     (strip-extension pathname)
				     "_"
				     ((case mode
				       ((forward) tapenade-forward-suffix)
				       ((reverse) tapenade-reverse-suffix)
				       (else (internal-error)))
				      i))
				    "f")))
				 subprograms-needing-AD)))
			      tapenade-path
			      tapenade-path)))
		(when (or show-commands? show-all?) (display command) (newline))
		(system command))
	       (unless keep?
		(for-each-indexed
		 (lambda (subprogram i)
		  (let ((mode (cond ((forward? (first subprogram)) 'forward)
				    ((reverse? (first subprogram)) 'reverse)
				    (else (internal-error)))))
		   (rm (replace-extension
			(string-append (strip-extension pathname)
				       "_"
				       ((case mode
					 ((forward) tapenade-forward-suffix)
					 ((reverse) tapenade-reverse-suffix)
					 (else (internal-error)))
					i))
			"f"))
		   (rm (replace-extension
			(string-append (strip-extension pathname)
				       "_"
				       ((case mode
					 ((forward) tapenade-forward-suffix)
					 ((reverse) tapenade-reverse-suffix)
					 (else (internal-error)))
					i))
			"msg"))))
		 subprograms-needing-AD)
		(rm (replace-extension pathname "f"))))))
	    (cond (adifor? (adifor))
		  (tapenade? (tapenade))
		  (no-AD? (write-file (unparse-blocks blocks)
				      (replace-extension pathname "f")))
		  ;; The default
		  (else (tapenade)))))))))

;;; Tam V'Nishlam Shevah L'El Borei Olam

;;; Local Variables:
;;; lisp-body-indent: 1
;;; End:
